"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import gffutils

# Gene ID identifiers
id_spec = {'gene': ['gene_id', 'geneID', 'ID=gene:']}


def CreateDB(FileNameGFF, FileNameDB, gene_features):
    """Create a GFF database using GFFUtils."""
    # Read the annotation file
    with open(FileNameGFF, 'r') as f:
        lines = f.readlines()

    # Keep only lines describing genes as features
    genes = '\n'.join([
        '\t'.join(line.split('\t')[:2] + ['gene'] + line.split('\t')[3:])
        for line in lines
        if (line[0] != '#' and line.split('\t')[2] in gene_features)])

    # Create the DB
    gffutils.create_db(
        data=genes, from_string=True, dbfn=FileNameDB, force=True,
        keep_order=True, merge_strategy='create_unique',
        disable_infer_transcripts=True, disable_infer_genes=True,
        sort_attribute_values=True, id_spec=id_spec)



"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy.optimize import fmin_tnc
from scipy.special import logsumexp

from omniCLIP.omni_stat import FitBinoDirchEmmisionProbabilities as FBProb


def pred_log_lik(counts, state, EmissionParameters, single_mix=None):
    """Compute the log_likelihood for counts."""
    params = EmissionParameters['Diag_event_params']
    alpha = params['alpha'][state]

    # Selecting the appropratie function
    if EmissionParameters['diag_event_mod'] == 'DirchMult':
        fitFun = FBProb.ComputeStateProbForGeneMD_unif
    elif EmissionParameters['diag_event_mod'] == 'DirchMultK':
        fitFun = FBProb.ComputeStateProbForGeneMD_unif_rep
    else:
        return None

    # Check which function to use for prediction the log-likelihood
    if single_mix is None:
        if params['nr_mix_comp'] > 1:  # Check for multiple mix components
            # Iterate over the mixtures and sum up the probabilities
            Prob = np.zeros((params['nr_mix_comp'], counts.shape[1]))
            for curr_mix_comp in range(0, params['nr_mix_comp']):
                # Compute the MultDirch component of the probability
                Prob[curr_mix_comp, :] = fitFun(
                    counts, alpha[:, curr_mix_comp],
                    state, EmissionParameters)

                # Compute the mixture component
                Prob[curr_mix_comp, :] += np.log(
                    params['mix_comp'][state][curr_mix_comp])

            # Sum the probabilities
            Prob = logsumexp(Prob, axis=0)

        else:
            Prob = fitFun(counts, alpha[:, 0], state, EmissionParameters)

    else:  # Extract a single component from the mixture
        Prob = fitFun(counts, alpha[:, single_mix], state, EmissionParameters)

    return Prob


def estimate_multinomial_parameters(Counts, NrOfCounts,
                                    EmissionParameters, x_0):
    """Estimate the DirchMult parameters of a mixture component."""
    if len(Counts.shape) == 1:
        Counts = np.expand_dims(Counts, axis=1)
    args_TC = (Counts, NrOfCounts, EmissionParameters)
    Bounds = tuple([(1e-100, None) for i in range(0, len(x_0))])
    disp = 1 if EmissionParameters['verbosity'] > 0 else 0

    if EmissionParameters['diag_event_mod'] == 'DirchMult':
        fitFun = FBProb.MD_f_joint_vect_unif
    elif EmissionParameters['diag_event_mod'] == 'DirchMultK':
        fitFun = FBProb.MDK_f_joint_vect_unif
    else:
        return None

    alpha = fmin_tnc(
        fitFun, x_0, fprime=FBProb.MD_f_prime_joint_vect_unif,
        args=args_TC, bounds=Bounds, disp=disp, maxfun=50)[0]

    return alpha



"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import random
from scipy.special import logsumexp
from sklearn.linear_model import SGDClassifier
import sys
import time

from omniCLIP.data_parsing import tools
from omniCLIP.omni_stat.utils import get_mem_usage


def PredictTransistions(Counts, TransitionParameters, NrOfStates, Type='multi',
                        verbosity=1):
    """Predict the transition probabilities for a gene."""
    return PredictTransistionsSimple(Counts, TransitionParameters, NrOfStates)


def PredictTransistionsSimple(Counts, TransitionParameters, NrOfStates,
                              verbosity=1):
    """Predict the transition probabilities for a gene."""
    TransitionParametersLogReg = TransitionParameters[1]
    TransistionProb = (np.ones((NrOfStates, NrOfStates, Counts.shape[1]))
                       * np.log((1 / np.float64(NrOfStates))))

    # Genererate the features
    CovMat = GenerateFeatures(
        np.array(list(range(Counts.shape[1] - 1))), Counts)

    ix_nonzero = np.sum(CovMat, axis=0) > 0
    # Create the probabilities
    TempProb = TransitionParametersLogReg.predict_log_proba(CovMat.T).T

    NormFactor = np.ones((TempProb.shape[1]))
    for CurrentState in range(NrOfStates):
        for NextState in range(NrOfStates):
            if CurrentState == NextState:
                TransistionProb[NextState, CurrentState, 1:] = TempProb[1, :]
            else:
                TransistionProb[NextState, CurrentState, 1:] = TempProb[0, :]

        # Normalize the transition probabilities

        if CurrentState == 0:
            # Only compute the normalizing factor once as compuation is
            # expensive
            if np.sum(ix_nonzero) > 0:
                # Nonzero entries
                NormFactor[ix_nonzero] = logsumexp(
                    TransistionProb[:, CurrentState, 1:][:, ix_nonzero],
                    axis=0)
            if np.sum(ix_nonzero == 0) > 0:
                # Zero entries
                first_zero_pos = np.where(ix_nonzero == 0)[0][0]
                NormFactor[ix_nonzero == 0] = logsumexp(
                    TransistionProb[:, CurrentState, 1:][:, first_zero_pos],
                    axis=0)

        for NextState in range(NrOfStates):
            TransistionProb[NextState, CurrentState, 1:] -= NormFactor

    del TempProb
    return TransistionProb


def FitTransistionParameters(Sequences, Background, TransitionParameters,
                             CurrPath, Type='multi', verbosity=1):
    """Determine optimal logistic regression parameters.

    Return the optimal parameters of the logistic regression for predicting
    the TransitionParameters.
    """
    print('Fitting transition parameters')
    get_mem_usage(verbosity)

    NewTransitionParametersLogReg = FitTransistionParametersSimple(
        Sequences, Background,
        TransitionParameters, CurrPath,
        verbosity=verbosity)

    get_mem_usage(verbosity)

    return NewTransitionParametersLogReg


def FitTransistionParametersSimple(Sequences, Background, TransitionParameters,
                                   CurrPath, verbosity=1):
    """Determine optimal logistic regression parameters.

    Return the optimal parameters of the logistic regression for predicting
    the TransitionParameters.
    """
    # Generate features from the CurrPaths and the Information in the coverage
    TransitionMatrix = TransitionParameters[0]
    NewTransitionParametersLogReg = {}
    t = time.time()

    # Iterate over the possible transitions
    assert (TransitionMatrix.shape[0] > 1), 'Only two states are currently allowed'

    genes = list(CurrPath.keys())
    genes = random.sample(genes, min(len(genes), 1000))

    NrOfStates = TransitionMatrix.shape[0]
    Xs = []
    Ys = []
    SampleSame = []
    SampleOther = []
    print("Learning transition model")
    print("Iterating over genes")
    get_mem_usage(verbosity, msg='Fitting transition parameters: I')

    for i, gene in enumerate(genes):
        if i % 1000 == 0:
            sys.stdout.write('.')
            sys.stdout.flush()
        # Get data
        Sequences_per_gene = tools.PreloadSequencesForGene(Sequences, gene)
        CovMat = tools.StackData(Sequences_per_gene, add='all')
        CovMat[CovMat < 0] = 0
        nr_of_samples = CovMat.shape[0]
        for CurrState in range(NrOfStates):
            for NextState in range(NrOfStates):
                # Positions where the path is in the current state
                Ix1 = CurrPath[gene][:-1] == CurrState
                # Positions where the subsequent position path is in the "next"
                # state
                Ix2 = CurrPath[gene][1:] == NextState
                # Positions where the path changes from the current state to
                # the other state
                Ix = np.where(Ix1 * Ix2)[0]

                CovMatIx = GenerateFeatures(Ix, CovMat)

                if CurrState == NextState:
                    if CovMatIx.shape[1] == 0:
                        CovMatIx = np.zeros((nr_of_samples, 1))
                        SampleSame.append(CovMatIx)
                    else:
                        SampleSame.append(CovMatIx)
                else:
                    if CovMatIx.shape[1] == 0:
                        CovMatIx = np.zeros((nr_of_samples, 1))
                        SampleOther.append(CovMatIx)
                    else:
                        SampleOther.append(CovMatIx)
        del Sequences_per_gene, CovMat

    get_mem_usage(verbosity, msg='Fitting transition parameters: II')

    len_same = np.sum([Mat.shape[1] for Mat in SampleSame])
    len_other = np.sum([Mat.shape[1] for Mat in SampleOther])

    X = np.concatenate(SampleSame + SampleOther, axis=1).T
    del SampleSame, SampleOther

    # Create Y
    Y = np.hstack(
        (np.ones((1, len_same), dtype=np.int),
         np.zeros((1, len_other), dtype=np.int)))[0, :].T
    classes = np.unique(Y)

    get_mem_usage(verbosity, msg='Fitting transition parameters: III')
    n_iter = max(5, np.ceil(10**6 / Y.shape[0]))

    NewTransitionParametersLogReg = SGDClassifier(loss="log", max_iter=n_iter)
    ix_shuffle = np.arange(X.shape[0])
    for n in range(n_iter):
        np.random.shuffle(ix_shuffle)
        for batch_ix in np.array_split(ix_shuffle, 50):
            NewTransitionParametersLogReg.partial_fit(
                X[batch_ix, :], Y[batch_ix], classes=classes)

    del Ix1, Ix2,  Ix, X, Y, Xs, Ys
    get_mem_usage(verbosity, t=t, msg='Fitting transition parameters: IV')

    return NewTransitionParametersLogReg


def GenerateFeatures(Ix, CovMat):
    """Generate the coverage matrix features for the logistic regression."""
    FeatureMatrix = np.log(1 + CovMat[:, Ix])
    return FeatureMatrix



"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from copy import deepcopy
import itertools
import multiprocessing
import numpy as np
from scipy.special import logsumexp

from omniCLIP.omni_stat import diag_event_model
from omniCLIP.omni_stat import emission_prob


def em(counts, nr_of_counts, EmissionParameters, x_0=None, First=False, max_nr_iter=15, tol=0.0001, rand_sample_size=10, verbosity=1):
    """Perform the EM algorithm."""
    template_state = 3
    fg_state, bg_state = emission_prob.get_fg_and_bck_state(EmissionParameters, final_pred=True)
    check = False

    OldEmissionParameters = deepcopy(EmissionParameters)
    for curr_state in list(counts.keys()):
        # Only compute the the emission probabilities once
        if EmissionParameters['diag_bg']:
            if curr_state != fg_state:
                if True:
                    if check is True:
                        print('Using template state ' + str(curr_state))
                        EmissionParameters['Diag_event_params']['mix_comp'][curr_state] = deepcopy(EmissionParameters['Diag_event_params']['mix_comp'][template_state])
                        EmissionParameters['Diag_event_params']['alpha'][curr_state] = deepcopy(EmissionParameters['Diag_event_params']['alpha'][template_state])
                        continue
                    else:
                        print('setting template state ' + str(curr_state))
                        check = True
                        template_state = curr_state
                else:
                    template_state = 3
                    check = True
                    EmissionParameters['Diag_event_params']['mix_comp'][curr_state] = deepcopy(EmissionParameters['Diag_event_params']['mix_comp'][template_state])
                    EmissionParameters['Diag_event_params']['alpha'][curr_state] = deepcopy(EmissionParameters['Diag_event_params']['alpha'][template_state])
                    continue
        print('Estimating state ' + str(curr_state))

        curr_counts = counts[curr_state]
        curr_nr_of_counts = nr_of_counts[curr_state]

        alpha, mixtures = Parallel_estimate_mixture_params(OldEmissionParameters, curr_counts, curr_nr_of_counts, curr_state, rand_sample_size, max_nr_iter, nr_of_iter=20, stop_crit=1.0, nr_of_init=10, verbosity=verbosity)
        EmissionParameters['Diag_event_params']['alpha'][curr_state] = alpha
        EmissionParameters['Diag_event_params']['mix_comp'][curr_state] = mixtures

    return EmissionParameters


def Parallel_estimate_mixture_params(EmissionParameters, curr_counts_orig, curr_nr_of_counts_orig, curr_state, rand_sample_size, max_nr_iter, nr_of_iter=20, stop_crit=1.0, nr_of_init=10, verbosity=1):
    """Estimate the dirichlet multinomial mixture parameters."""
    # 1) Copy old parameters and use it as initialisation for the first iteration
    alphas_list = []
    mixtures_list = []
    lls_list = []
    curr_counts = deepcopy(curr_counts_orig)
    curr_nr_of_counts = deepcopy(curr_nr_of_counts_orig)

    if len(curr_counts.shape) == 1:
        curr_counts = np.expand_dims(curr_counts, axis=1)

    if np.sum(np.sum(curr_counts, axis=0) > 0) > 0:
        curr_nr_of_counts = curr_nr_of_counts[:, np.sum(curr_counts, axis=0) >0]
        curr_counts = curr_counts[:, np.sum(curr_counts, axis=0) >0]

    # Test for fitting distributions only on diag events
    if np.sum(np.sum(curr_counts, axis=0) > 10) > 10:
        curr_nr_of_counts = curr_nr_of_counts[:, np.sum(curr_counts, axis=0) > 10]
        curr_counts = curr_counts[:, np.sum(curr_counts, axis=0) > 10]

    tracks_per_rep = EmissionParameters['Diag_event_params']['alpha'][curr_state].shape[0]
    NrOfReplicates = curr_counts.shape[0] // tracks_per_rep

    if len(curr_counts.shape) == 1:
        curr_counts = np.expand_dims(curr_counts, axis=1)

    # Save old lls mixtures and alphas
    mixtures = deepcopy(EmissionParameters['Diag_event_params']['mix_comp'][curr_state])

    scored_counts = score_counts(curr_counts, curr_state, EmissionParameters)
    scored_counts += np.tile(np.log(mixtures[:, np.newaxis]), (1, scored_counts.shape[1]))
    ll = np.sum(np.sum(logsumexp(scored_counts, axis=0) + np.log(curr_nr_of_counts)))

    alphas_list.append(deepcopy(EmissionParameters['Diag_event_params']['alpha'][curr_state]))
    mixtures_list.append(deepcopy(EmissionParameters['Diag_event_params']['mix_comp'][curr_state]))
    lls_list.append(ll)

    data = zip(itertools.repeat(stop_crit), itertools.repeat(rand_sample_size), itertools.repeat(max_nr_iter), list(range(nr_of_init)), itertools.repeat(EmissionParameters), itertools.repeat(curr_state), itertools.repeat(curr_counts), itertools.repeat(curr_nr_of_counts)   )

    if EmissionParameters['nb_proc'] == 1:
        results = [Parallel_estimate_single_mixture_params(args) for args in data]
    else:
        print("Spawning processes")
        pool = multiprocessing.get_context("spawn").Pool(EmissionParameters['nb_proc'], maxtasksperchild=5)
        results = pool.imap(Parallel_estimate_single_mixture_params, data, chunksize=1)
        pool.close()
        pool.join()
        print("Collecting results")
        results = [res for res in results]

    alphas_list += [res[0] for res in results]
    mixtures_list += [res[1] for res in results]
    lls_list += [res[2] for res in results]

    # Select which alpha had the highest ll
    max_ll_pos = np.argmax(np.array(lls_list))

    alpha = alphas_list[max_ll_pos]
    mixtures = mixtures_list[max_ll_pos]
    return alpha, mixtures


def Parallel_estimate_single_mixture_params(args):
    """Estimate thedirichlet multinomial mixture parameters."""

    stop_crit, rand_sample_size, max_nr_iter, curr_init, EmissionParameters, curr_state, curr_counts, curr_nr_of_counts = args
    # Compute the curr mixture, ll and alpha
    # Initialiste the parameters
    old_ll = 0
    ll = -10

    OldAlpha = deepcopy(EmissionParameters['Diag_event_params']['alpha'][curr_state])
    mixtures = deepcopy(EmissionParameters['Diag_event_params']['mix_comp'][curr_state])

    if curr_init > 0:
        OldAlpha = np.random.uniform(low=0.0001, high=0.1, size=OldAlpha.shape)
        for i in range(OldAlpha.shape[1]):
            OldAlpha[np.random.randint(OldAlpha.shape[0]-1), i] = np.random.random() * 10.0
            OldAlpha[-2, i] = np.random.random() * 1.0
            OldAlpha[-1, i] = np.random.random() * 10.0
        mixtures = np.random.uniform(low=0.0001, high=1.0, size=mixtures.shape)
        mixtures /= np.sum(mixtures)
    if EmissionParameters['Diag_event_params']['nr_mix_comp'] == 1:
        # Case that only one mixture component is given
        EmissionParameters['Diag_event_params']['alpha'][curr_state][:, 0] = diag_event_model.estimate_multinomial_parameters(curr_counts, curr_nr_of_counts, EmissionParameters, OldAlpha[:])
        # Compute ll
        scored_counts = score_counts(curr_counts, curr_state, EmissionParameters)
        scored_counts += np.tile(np.log(mixtures[:, np.newaxis]), (1, scored_counts.shape[1]))
        ll = np.sum(np.sum(logsumexp(scored_counts, axis=0) + np.log(curr_nr_of_counts)))

        OldAlpha = deepcopy(EmissionParameters['Diag_event_params']['alpha'][curr_state])
        mixtures = deepcopy(EmissionParameters['Diag_event_params']['mix_comp'][curr_state])

    else:
        zero_ix = []
        for iter_nr in range(max_nr_iter):
            print('em-iteration ' + str(iter_nr))

            scored_counts = score_counts(curr_counts, curr_state, EmissionParameters)
            scored_counts += np.tile(np.log(mixtures[:, np.newaxis]), (1, scored_counts.shape[1]))
            # 2) Compute the mixture components
            # Compute the normalisation factor
            normalised_likelihood = logsumexp(scored_counts, axis=0)

            old_ll = ll
            ll = np.sum(np.sum(logsumexp(scored_counts, axis=0) + np.log(curr_nr_of_counts)))

            if np.abs(old_ll - ll) < stop_crit:
                if len(zero_ix) == 0:
                    break

            normalised_scores = scored_counts - np.tile(normalised_likelihood, (scored_counts.shape[0], 1))
            un_norm_mixtures = logsumexp(normalised_scores, b=np.tile(curr_nr_of_counts, (scored_counts.shape[0], 1)), axis = 1)

            mixtures = np.exp(un_norm_mixtures - logsumexp(un_norm_mixtures))

            # 3) Compute for eachcount the most likely mixture component
            curr_weights = np.exp(normalised_scores)
            curr_weights = (curr_weights == np.tile(np.max(curr_weights, axis=0), (curr_weights.shape[0], 1))) *1.0

            zero_mix = np.sum(curr_weights, axis=1) == 0
            zero_ix = np.where(zero_mix)[0].tolist()

            EmissionParameters['Diag_event_params']['mix_comp'][curr_state] = mixtures
            # Get number of positions that are used. (In case there are fewer entries that rand_sample_size in counts)
            rand_size = min(rand_sample_size, curr_counts.shape[1])
            for i in zero_ix:
                random_ix = np.random.choice(curr_counts.shape[1], rand_size, p=(curr_nr_of_counts[0, :] / np.float(np.sum(curr_nr_of_counts[0, :]))))
                curr_counts = np.hstack([curr_counts, curr_counts[:, random_ix]])
                curr_nr_of_counts = np.hstack([curr_nr_of_counts, np.ones((1, rand_size))])
                temp_array = np.zeros((normalised_scores.shape[0], rand_size))
                temp_array[i, :] = i
                normalised_scores = np.hstack([normalised_scores, temp_array])
                temp_array = np.zeros((curr_weights.shape[0], rand_size))
                temp_array[i, :] = 1
                curr_weights = np.hstack([curr_weights, temp_array])

            # 4) Compute the dirichlet-multinomial parameters
            for curr_mix_comp in range(EmissionParameters['Diag_event_params']['nr_mix_comp']):
                local_counts = curr_counts
                local_nr_counts = curr_nr_of_counts * curr_weights[curr_mix_comp, :]
                local_counts = local_counts[:, local_nr_counts[0, :] > 0]
                local_nr_counts = local_nr_counts[0, local_nr_counts[0, :] > 0]
                if len(local_counts.shape) == 1:
                    local_counts = np.expand_dims(local_counts, axis=1)
                curr_alpha = diag_event_model.estimate_multinomial_parameters(local_counts, local_nr_counts, EmissionParameters, OldAlpha[:, curr_mix_comp])

                if curr_mix_comp in zero_ix:
                    OldAlpha[:, curr_mix_comp] = np.random.uniform(low=0.0001, high=0.1, size=OldAlpha[:, curr_mix_comp].shape)
                    OldAlpha[np.random.randint(OldAlpha.shape[0]), curr_mix_comp] = np.random.random() * 10.0
                    OldAlpha[-2, curr_mix_comp] = np.random.random() * 1.0
                    OldAlpha[-1, curr_mix_comp] = np.random.random() * 10.0
                else:
                    OldAlpha[:, curr_mix_comp] = curr_alpha

            if (len(zero_ix) > 0) and (iter_nr + 2 < max_nr_iter):
                # Treat the case where some mixtures have prob zero
                mixtures[zero_ix] = np.mean(mixtures)
                mixtures /= np.sum(mixtures)
                EmissionParameters['Diag_event_params']['mix_comp'][curr_state] = deepcopy(mixtures)
            EmissionParameters['Diag_event_params']['alpha'][curr_state] = deepcopy(OldAlpha)
            # Check if convergence has been achieved.

        mixtures[zero_ix] = np.min(mixtures[mixtures > 0])
        mixtures /= np.sum(mixtures)

    return [deepcopy(OldAlpha), mixtures, ll]


def score_counts(counts, state, EmissionParameters):
    """Scores the counts for each mixture component."""

    nr_mixture_components = EmissionParameters['Diag_event_params']['nr_mix_comp']
    # Initialize the return array
    scored_counts = np.zeros((nr_mixture_components, counts.shape[1]))

    # Compute for each state the log-likelihood of the counts
    for mix_comp in range(nr_mixture_components):
        scored_counts[mix_comp, :] = diag_event_model.pred_log_lik(counts, state, EmissionParameters, single_mix=mix_comp)
        scored_counts[mix_comp, :] += np.log(EmissionParameters['Diag_event_params']['mix_comp'][state][mix_comp])

    return scored_counts



"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import re
import warnings


def GetRawCoverageFromRegion(
        SamReader, Chr, Start, Stop, Collapse=False, CovType='coverage',
        Genome='', legacy=True, mask_flank_variants=3, max_mm=2,
        rev_strand=None, ign_out_rds=False, gene_strand=0):
    """Extract coverage from a BAM-file.

    This function gets from a BAM-file the coverage and returns it as a sparse
    vector for each chromosome and strand.
    """
    # Initiate nucleotide lookup
    NuclDict = {'A': 0, 'C': 1, 'G': 2, 'T': 3, 'D': 4}

    # Prepare regular expression
    r = re.compile('([\\^]*[ACGTN]+)[0]*')

    # Compute Length of the region
    RegionLength = Stop - Start

    # Initialise the length vectors
    ret_arrays = dict()
    ret_arrays['variants'] = np.zeros((5, RegionLength), dtype=np.int32)
    ret_arrays['read-ends'] = np.zeros((2, RegionLength), dtype=np.int32)
    ret_arrays['coverage'] = np.zeros((1, RegionLength), dtype=np.int32)

    # Modify gene_strand if the reads are coming from the reverse strand
    if rev_strand is not None:
        if rev_strand == 0:
            gene_strand *= -1  # Swap the strand

    # Iterate over Reads
    iter = SamReader.fetch(Chr, Start, Stop)
    for CurrRead in iter:
        # Check for mismatches
        if CurrRead.get_tag('NM') > max_mm:
            continue

        # Check for position whithin gene boundaries
        CurrReadstart = CurrRead.pos - Start
        if ign_out_rds:
            LastPos = (CurrReadstart
                       + sum([e[1] for e in CurrRead.cigar if e[0] != 4])
                       - 1)

            if (CurrReadstart < 0) or (LastPos > Stop - Start):
                continue

        # Check whether the read strand matches else skip this read
        if rev_strand is not None:
            # Check if read is paired
            if CurrRead.flag & 1:
                if gene_strand == -1:
                    # CurrRead.flag & 16 means read is on reverse strand
                    # CurrRead.flag & 32 means mate read is on reverse strand
                    # CurrRead.flag & 64 means read is first in pair
                    # CurrRead.flag & 128 means read is second in pair
                    if ((CurrRead.flag & 16) > 0) & ((CurrRead.flag & 64) > 0):
                        continue
                    if ((CurrRead.flag & 16) == 0) & ((CurrRead.flag & 128) > 0):
                        continue
                else:
                    if ((CurrRead.flag & 16) > 0) & ((CurrRead.flag & 128) > 0):
                        continue
                    if ((CurrRead.flag & 16) == 0) & ((CurrRead.flag & 64) > 0):
                        continue
            else:
                if gene_strand == -1:
                    if ((CurrRead.flag & 16) > 0):
                        continue
                else:
                    if ((CurrRead.flag & 16) == 0):
                        continue

        # Compute relative positions of read
        CurrReadstart = CurrRead.pos - Start
        FirstPos = CurrReadstart
        LastPos = FirstPos + sum([e[1] for e in CurrRead.cigar if e[0] != 4])
        # Check ho many reads the current read represents (Collapsing)
        if Collapse:
            Mult = CurrRead.qname.split('-')
            if len(Mult) == 1:
                raise Exception('Error: Collapsing of read: ' + CurrRead.qname)
            Mult = int(Mult[-1])
        else:
            Mult = 1

        # Processing for variants
        if 'variants' in CovType:
            GlobalVariantPos = GetVariantsFromRead(CurrRead, r)
            # Transform letters into numbers, A - 0, C - 1, G - 2, T - 3
            for e in GlobalVariantPos:
                if e[1] == 'N':
                    continue
                # Check if variant falls into flanks
                if (e[0] - Start) < (FirstPos + mask_flank_variants):
                    continue
                elif (e[0] - Start) > (LastPos - 1 - mask_flank_variants):
                    continue
                # Check whether the variant lies outside of the gene
                if e[0] - Start < 0 or e[0] - Start >= RegionLength:
                    continue
                # Process the variant
                ret_arrays['variants'][NuclDict[e[1]], e[0] - Start] += Mult

        # Processing for read extremities
        if 'read-ends' in CovType:
            if CurrRead.flag & 1:
                # Check if the read is paired end.
                # If yes chose the outer ends as ends.
                if FirstPos >= 0:
                    if ((CurrRead.flag & 128) > 0) & ((CurrRead.flag & 32) > 0):
                        ret_arrays['read-ends'][0, CurrReadstart] += Mult
                    if ((CurrRead.flag & 64) > 0) & ((CurrRead.flag & 32) > 0):
                        ret_arrays['read-ends'][1, CurrReadstart] += Mult
                if LastPos < RegionLength:
                    if ((CurrRead.flag & 16) > 0) & ((CurrRead.flag & 64) > 0):
                        ret_arrays['read-ends'][1, LastPos - 1] += Mult
                    if ((CurrRead.flag & 16) > 0) & ((CurrRead.flag & 128) > 0):
                        ret_arrays['read-ends'][0, LastPos - 1] += Mult
            else:
                if FirstPos >= 0:
                    if ((CurrRead.flag & 16) > 0):
                        ret_arrays['read-ends'][1, CurrReadstart] += Mult
                    else:
                        ret_arrays['read-ends'][0, CurrReadstart] += Mult
                if LastPos < RegionLength:
                    if ((CurrRead.flag & 16) > 0):
                        ret_arrays['read-ends'][0, LastPos - 1] += Mult
                    else:
                        ret_arrays['read-ends'][1, LastPos] += Mult

        # Processing for coverage
        for cig in CurrRead.cigar:
            # Check which type to get the coverage for
            if cig[0] == 0:
                _start = max(0, CurrReadstart)
                _end = min(RegionLength, max(0, CurrReadstart + cig[1]))
                ret_arrays['coverage'][0, _start:_end] += Mult
            if cig[0] != 4:
                CurrReadstart += cig[1]

    return ret_arrays


def GetVariantsFromRead(CurrRead, r):
    """Parse variants from read.

    Takes a pySAM read and returns variants based on the MD Tag the Variants
    and their absolute positions.
    """
    # Get the sequence
    Seq = CurrRead.seq

    # Get the MD tag
    Tag = CurrRead.get_tag('MD')

    # Split the string
    SplitTag = [e for e in r.split(Tag) if len(e) > 0]
    if len(SplitTag) == 1:
        return []

    # Convert the groups to a list of local positions and nucleotides
    TempPos = 0
    Pos = []
    for i in range(0, len(SplitTag)):
        if SplitTag[i].isdigit():
            # Increase the counter by the number of positions where there is no
            # mismatch
            TempPos += int(SplitTag[i])
        else:
            if SplitTag[i][0] == '^':  # Check if it is a deletion
                continue
            else:
                for l in range(len(SplitTag[i])):
                    Pos.append([TempPos, Seq[TempPos]])
                    TempPos += 1

    # Convert the local positions from Pos to global positions.
    CurrGlobalPos = CurrRead.pos
    GlobalPos = []

    # Iterate over the segments of the cigar string
    for cig in CurrRead.cigar:
        # Split the positions in Pos into the ones faling into the current
        # CIGAR segement and the rest

        # Sequence match
        if cig[0] == 0:
            CurrSeg = [e for e in Pos if e[0] < cig[1]]
            Pos = [[e[0] - cig[1], e[1]] for e in Pos if e[0] >= cig[1]]
            for e in CurrSeg:
                GlobalPos.append([CurrGlobalPos + e[0], e[1]])
            CurrGlobalPos += cig[1]

        # Insertion to the reference
        elif cig[0] == 1:
            continue

        # Deletion from the reference
        elif cig[0] == 2:
            for temp_pos in range(cig[1]):
                GlobalPos.append([CurrGlobalPos + temp_pos, 'D'])
            CurrGlobalPos += cig[1]
            continue

        # Skipped region from the reference
        elif cig[0] == 3:
            CurrGlobalPos += cig[1]

        # Soft clipping (clipped sequences present in SEQ)
        elif cig[0] == 4:
            continue

        # Hard clipping (clipped sequences NOT present in SEQ)
        elif cig[0] == 5:
            continue
            CurrGlobalPos += cig[1]

        # Padding (silent deletion from padded reference)
        elif cig[0] == 6:
            continue

        # Sequence match
        elif cig[0] == 7:
            CurrSeg = [e for e in Pos if e[0] < cig[1]]
            Pos = [[e[0] - cig[1], e[1]] for e in Pos if e[0] >= cig[1]]
            for e in CurrSeg:
                GlobalPos.append([CurrGlobalPos + e[0], e[1]])
            CurrGlobalPos += cig[1]

        # Sequence mismatch
        elif cig[0] == 8:
            CurrSeg = [e for e in Pos if e[0] < cig[1]]
            Pos = [[e[0] - cig[1], e[1]] for e in Pos if e[0] >= cig[1]]
            for e in CurrSeg:
                GlobalPos.append([CurrGlobalPos + e[0], e[1]])
            CurrGlobalPos += cig[1]

        else:
            warnings.warn("Encountered unhandled CIGAR character in read "
                          + CurrRead.qname)
            CurrGlobalPos += cig[1]

    return GlobalPos



#!/usr/bin/env python2.7

#"""
#    omniCLIP is a CLIP-Seq peak caller
#
#    Copyright (C) 2017 Philipp Boss#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#"""


cimport cython
from cython.view cimport array as cvarray
import numpy as np
cimport numpy as np

DTYPE = np.float64
ctypedef np.float64_t DTYPE_t

@cython.boundscheck(False)
def viterbi(np.ndarray[DTYPE_t, ndim=2] EmmisionProbabilites, np.ndarray[DTYPE_t, ndim=3] TransistionProbabilities, np.ndarray[DTYPE_t, ndim=2] PriorProbabilities):
    """
    This function determines the most likely path using the Viterbi algorithm
    """
    cdef int i, j, k, CurrArgmax
    cdef int [:,:] TraceBack
    cdef unsigned int SeqLen = EmmisionProbabilites.shape[1]
    cdef unsigned int NrOfStates = EmmisionProbabilites.shape[0]
    cdef np.ndarray[np.int_t, ndim=1] Path = np.zeros(SeqLen, dtype=np.int)
    cdef double [:,:] V
    cdef double CurrMax, Temp
    cdef LogLik = np.float64
    cdef unsigned int D_SIZE = sizeof(double)
    cdef unsigned int I_SIZE = sizeof(int)

    TraceBack = cvarray(shape=(NrOfStates, SeqLen + 1), itemsize=I_SIZE, format='i', mode="c")

    V = cvarray(shape=(NrOfStates, SeqLen + 1), itemsize=D_SIZE, format='d', mode="c")

    #First step of Viterbi
    for j in xrange(NrOfStates):
        V[j, 0] = EmmisionProbabilites[j, 0] + PriorProbabilities[j, 0]

    #Iterate over the positions
    for i in xrange(0, SeqLen):
        for j in xrange(NrOfStates):
            CurrArgmax = 0
            CurrMax = V[0, i] + TransistionProbabilities[0, j, i]
            for k in xrange(1, NrOfStates):
                Temp = V[k, i] + TransistionProbabilities[k, j, i]
                if CurrMax < Temp:
                    CurrMax = Temp
                    CurrArgmax = k

            TraceBack[j, i] = CurrArgmax
            V[j, i + 1] = V[CurrArgmax, i] + EmmisionProbabilites[j, i] + TransistionProbabilities[CurrArgmax, j, i]

    #Perform traceback
    CurrArgmax = 0
    CurrMax = V[0, SeqLen]
    for k in xrange(1, NrOfStates):
        Temp = V[k, SeqLen]
        if CurrMax < Temp:
            CurrMax = Temp
            CurrArgmax = k

    k = CurrArgmax
    Path[SeqLen - 1] = k
    for i in xrange(SeqLen - 2, -1, -1):
        k = TraceBack[k, i]
        Path[i] = k

    LogLik = V[0, SeqLen]

    return Path, LogLik



"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from Bio import SeqIO
from collections import defaultdict
from scipy.sparse import csr_matrix
import gzip
import h5py
import numpy as np
import os
import pysam

from omniCLIP.data_parsing import GetCoverageFromBam


def get_data_handle(file_path, write=False):
    """Return the h5py file handle."""
    if write:
        return h5py.File(file_path, 'r+')
    else:
        return h5py.File(file_path, 'r')


def close_data_handles(handle=False, handles=False):
    """Close all opened H5PY file handles."""
    if handle:
        handle.close()
    if handles:
        for _handle in handles:
            _handle.close()


def filter_gene_coverage(g_cove, min_coverage, min_peak):
    """Filter genes according to the coverage profile."""
    total_cov = sum([g_cove[rep][()].sum() for rep in list(g_cove.keys())])
    max_cov = np.max([g_cove[rep][()].max() for rep in list(g_cove.keys())])

    if (total_cov >= min_coverage) and (max_cov >= min_peak):
        return True
    else:
        return False


def load_data(bam_files, genome_dir, gene_annotation, out_file, Collapse=False,
              OnlyCoverage=False, mask_flank_variants=3, max_mm=2,
              ign_out_rds=False, rev_strand=None, CLIP_exp=False,
              min_coverage=100, min_peak=5):
    """Read the data from the bam-files."""
    SeqFile = h5py.File(out_file, 'w')

    if OnlyCoverage:
        print('Loading coverage only')

    # Defining vars
    trdict = {"A": 0, "C": 1, "G": 2, "T": 3, "N": 4}

    print("Parsing the gene annotation")
    Genes = [gene for gene in gene_annotation.features_of_type('gene')
             if '_PAR_Y' not in gene.id]

    # Create a dictionary of genes by chromosome
    genes_chr_dict = defaultdict(list)
    for gene in Genes:
        genes_chr_dict[gene.chrom].append(gene)

    # Create a list of cromosomes
    Chrs = [chrom for chrom in list(genes_chr_dict.keys())
            if chrom.lower() != 'chrm']

    # Creating the BAM files handle
    bam_handles = dict()
    for i, bam_file in enumerate(bam_files):
        bam_handles[i] = pysam.Samfile(bam_file, 'rb')

    # Identifying coverage to compute
    CovType = ['coverage']
    if not OnlyCoverage:
        CovType += ['variants', 'read-ends']

    # 3. Iterate over the CurrChromosomes
    for CurrChr in Chrs:
        print('Processing ' + CurrChr)
        # Get the genome
        CurrChrFile = os.path.join(genome_dir, CurrChr + '.fa.gz')
        if not os.path.isfile(CurrChrFile):
            print('Warning, chromosome not found: ' + CurrChrFile)
            continue
        with gzip.open(CurrChrFile, 'rt') as handle:
            record_dict = SeqIO.to_dict(SeqIO.parse(handle, 'fasta'))
            CurrChrSeq = record_dict[CurrChr]

        # Iterate over the Genes
        for gene in genes_chr_dict[CurrChr]:
            start = gene.start
            stop = gene.stop
            gene_id = gene.id
            strand = 1 if gene.strand == '+' else -1

            # Collect data for the different BAM files
            Coverage, Variants, ReadEnds = dict(), dict(), dict()
            for i in range(len(bam_files)):
                # Get the gene coverage
                ret_arrays = GetCoverageFromBam.GetRawCoverageFromRegion(
                    bam_handles[i], CurrChr, start, stop, Collapse=Collapse,
                    CovType=CovType, Genome='', legacy=False,
                    mask_flank_variants=mask_flank_variants,
                    max_mm=max_mm, ign_out_rds=ign_out_rds,
                    gene_strand=strand, rev_strand=rev_strand)
                Coverage[i] = ret_arrays['coverage']
                if not OnlyCoverage:
                    Variants[i] = ret_arrays['variants']
                    ReadEnds[i] = ret_arrays['read-ends']

            # Filter genes according to minimal CLIP peak
            is_peak = filter_gene_coverage(Coverage, min_coverage, min_peak)
            if CLIP_exp and (not is_peak):
                continue

            # Adding the gene to the SeqFile
            if gene_id not in SeqFile:
                SeqFile.create_group(gene_id)
                SeqFile[gene_id].create_dataset('strand', data=strand)
                SeqFile[gene_id].create_group('SummedCoverage')
                SeqFile[gene_id].create_group('Coverage')
            if not OnlyCoverage:
                SeqFile[gene_id].create_group('Variants')
                SeqFile[gene_id].create_group('Read-ends')

            # For all BAM files
            for i in range(len(bam_files)):
                # Save the unmodified coverage
                SeqFile[gene_id]['SummedCoverage'].create_dataset(
                    str(i), data=Coverage[i], compression="gzip",
                    compression_opts=9, chunks=Coverage[i].shape)

                if not OnlyCoverage:
                    Coverage[i] = (Coverage[i]
                                   - np.sum(Variants[i], axis=0)
                                   - np.sum(ReadEnds[i], axis=0))

                # Save the total coverage
                SeqFile[gene_id]['Coverage'].create_dataset(
                    str(i), data=Coverage[i], compression="gzip",
                    compression_opts=9, chunks=Coverage[i].shape)

                # Calculate the Variants and ReadEnds
                if not OnlyCoverage:
                    # Get the TC conversions
                    CurrSeq = str(CurrChrSeq.seq[start:stop]).upper()
                    GeneSeq = np.zeros((1, len(CurrSeq)), dtype=np.uint8)
                    GeneSeq[0, :] = np.array(
                        [trdict[e] for e in list(CurrSeq)])

                    # Ignore Variants at N positions
                    Variants[i][:, GeneSeq[0, :] == 4] = 0
                    Variants[i][Variants[i] < 0] = 0

                    non_zer_var = np.where(Variants[i])
                    ij = np.vstack((
                        GeneSeq[0, non_zer_var[1]] * 5 + non_zer_var[0],
                        non_zer_var[1]))
                    Variants_sparse = csr_matrix(
                        (Variants[i][np.where(Variants[i])], ij),
                        shape=(20, len(CurrSeq)))
                    Variants[i] = Variants_sparse.toarray()
                    del non_zer_var, ij

                    SeqFile[gene_id]['Variants'].create_group(str(i))
                    variant_fields = ['data', 'indptr', 'indices', 'shape']
                    for field in variant_fields:
                        SeqFile[gene_id]['Variants'][str(i)].create_dataset(
                            field, data=getattr(Variants_sparse, field))

                    # Only save the positions where a read in one of the
                    # replicates occured
                    SeqFile[gene_id]['Read-ends'].create_dataset(
                        str(i), data=ReadEnds[i], compression="gzip",
                        compression_opts=9)

    SeqFile.close()



"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import os
import random
import shutil

from omniCLIP.data_parsing import LoadReads
from omniCLIP.data_parsing import tools


def parsing_argparse(args):
    """Parse the argparse dict."""
    # Listing args to extract from the argparse parser
    named_args = [
        'bg_type', 'dat_file_bg', 'dat_file_clip', 'diag_bg', 'diag_event_mod',
        'emp_var', 'fg_pen', 'filter_snps', 'glm_weight', 'ign_diag',
        'ign_GLM', 'ign_out_rds', 'mask_ovrlp', 'max_it', 'nb_proc',
        'norm_class', 'out_dir', 'pv_cutoff', 'skip_diag_event_mdl',
        'snps_min_cov', 'snps_thresh', 'subs', 'tmp_dir', 'verbosity'
    ]

    # Creating the params dictionary
    params = {arg: getattr(args, arg) for arg in named_args}

    # Verifying the validity of the args
    verifying_args(params)

    # Printing args if verbosity
    if params['verbosity'] > 1:
        print(args)

    # Defining fixed-value params
    params['ExpressionParameters'] = [None, None]
    params['LastIter'] = False
    params['TransitionType'] = 'binary'

    # Conditional params - Number of States
    if params['bg_type'] == 'Coverage_bck':
        params['NrOfStates'] = 4
    else:
        params['NrOfStates'] = 3

    # Conditional params - Ignore reads outside gene and diag model
    if params['ign_out_rds']:
        params['ign_diag'] = params['ign_out_rds']

    # Conditional params - Prior Matrix
    params['PriorMatrix'] = (np.ones((params['NrOfStates'], 1))
                             / float(params['NrOfStates']))

    # Conditional  params - Transition Matrix
    TransMat = (np.ones((params['NrOfStates'], params['NrOfStates']))
                + np.eye(params['NrOfStates']))
    params['TransMat'] = TransMat / np.sum(np.sum(TransMat))

    # Conditional params - Diag Event Params, flag for penalized path pred
    params['Diag_event_params'] = {}
    params['Diag_event_params']['nr_mix_comp'] = args.nr_mix_comp
    params['Diag_event_params']['alpha'] = {}
    params['Diag_event_params']['mix_comp'] = {}
    for state in range(params['NrOfStates']):
        mixtures = np.random.uniform(0.0, 1.0, size=(args.nr_mix_comp))
        params['Diag_event_params']['mix_comp'][state] = (mixtures
                                                          / np.sum(mixtures))

    # Conditional params - out file name
    params['out_file_base'] = 'pred'
    if params['ign_GLM']:
        params['out_file_base'] += '_no_glm'
    if params['ign_diag']:
        params['out_file_base'] += '_no_diag'

    # Optional params - Random seeding
    if args.rnd_seed is not None:
        random.seed(args.rnd_seed)
        print('setting random seed')

    return params


def dup_seqfiles(params):
    """Create temporary Seqfiles that will be modified by omniCLIP.

    If no tmp_dir has been specified, both tmp files will be created in the
    same location as the CLIP dat file.
    """
    if params['tmp_dir'] is None:
        dir_path = os.path.dirname(os.path.realpath(params['dat_file_clip']))
        fg_tmp = params['dat_file_clip'] + '.tmp'
        bg_tmp = os.path.join(dir_path, 'bg_data.dat.tmp')
    else:
        fg_tmp = os.path.join(params['tmp_dir'], 'clip_data.dat.tmp')
        bg_tmp = os.path.join(params['tmp_dir'], 'bg_data.dat.tmp')

    shutil.copy(params['dat_file_clip'], fg_tmp)
    shutil.copy(params['dat_file_bg'], bg_tmp)

    # Open the temporary read files
    params['dat_file_clip'] = fg_tmp
    params['dat_file_bg'] = bg_tmp

    return params


def parsing_files(args, params):
    """Parse arguments that are function of the data files."""
    # Loading Sequence and Background
    Sequences = LoadReads.get_data_handle(params['dat_file_clip'], write=True)
    Background = LoadReads.get_data_handle(params['dat_file_bg'], write=True)

    # Estimate the library size
    params['LibrarySize'] = tools.estimate_library_size(Sequences)
    params['BckLibrarySize'] = tools.estimate_library_size(Background)

    # Estimate to number of replicates
    gene = list(Sequences.keys())[0]
    params['NrOfReplicates'] = len(Sequences[gene]['Coverage'])
    params['NrOfBckReplicates'] = len(Background[gene]['Coverage'])

    # Initialise the parameter vector alpha
    alphashape = (Sequences[gene]['Variants']['0']['shape'][0]
                  + Sequences[gene]['Coverage']['0'][()].shape[0]
                  + Sequences[gene]['Read-ends']['0'][()].shape[0])
    for state in range(params['NrOfStates']):
        params['Diag_event_params']['alpha'][state] = np.random.uniform(
            0.9, 1.1, size=(alphashape, args.nr_mix_comp))

    return params


def verifying_args(params):
    """Testing the arguments.

    TODO: Implement more testing."""
    pass



"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy.special import gammaln


def log_pdf_vect(k, alpha):
    """Compute the log pdf for an array of counts."""
    # First check whether the input is one dimensional and set parameters accordingly
    alpha = np.expand_dims(alpha, axis=1)
    if len(k.shape) == 1:
        Ks = 1
        k = np.expand_dims(k, axis=1)
    else:
        Ks = k.shape[1]
    Pos = gammaln(np.sum(k, axis=0) + 1) + np.tile(gammaln(np.sum(alpha)), (1, Ks)) + np.sum(gammaln(np.tile(alpha, (1, Ks)) + k), axis=0)
    Neg = np.sum(gammaln(k + 1), axis=0) + gammaln(np.sum(np.tile(alpha, (1, Ks)) + k, axis=0)) + np.tile(np.sum(gammaln(alpha)), (1, Ks))
    log = Pos - Neg

    return log


def expand_k(k):
    """Add one dimension to k if k is only a 1-dim  array."""
    if len(k.shape) == 1:
        Ks = 1
        k = np.expand_dims(k, axis=1)
    else:
        Ks = k.shape[1]
    return k, Ks


def log_pdf_vect_rep(Counts, alpha, tracks_per_rep, NrOfReplicates):
    """Compute the log pdf for an array of counts."""
    # First check whether the input is one dimensional and set parameters accordingly
    alpha = np.expand_dims(alpha, axis=1)

    # Compute the 'collapsed' counts per diagnostic events
    k = Counts[0:tracks_per_rep, :].copy()
    k, Ks = expand_k(k)
    for rep in range(1, NrOfReplicates):
        new_k, Ks = expand_k(Counts[rep * tracks_per_rep:(rep + 1) * tracks_per_rep, :])
        k += new_k

    # Compute the factors that are independent of the replicates
    Pos = np.sum(gammaln(k + np.tile(alpha, (1, Ks))), axis=0) + np.tile(gammaln(np.sum(alpha)), (1, Ks))
    Neg = np.tile(np.sum(gammaln(alpha)), (1, Ks)) + gammaln(np.sum(np.tile(alpha, (1, Ks)) + k, axis=0))

    for rep in range(NrOfReplicates):
        k = Counts[rep * tracks_per_rep:(rep + 1) * tracks_per_rep, :].copy()
        k, Ks = expand_k(k)
        Pos += gammaln(np.sum(k, axis=0) + 1)
        Neg += np.sum(gammaln(k + 1), axis=0)

    log = Pos - Neg

    return log


def TwoBinomlog_pdf_vect(k1, k2, alpha):
    """PDF for an array of counts with two multinomial distributions."""
    alpha = np.expand_dims(alpha, axis=1)
    if len(k1.shape) == 1:
        Ks = 1
        k1 = np.expand_dims(k1, axis=1)
        k2 = np.expand_dims(k2, axis=1)
    else:
        Ks = k1.shape[1]
    k = k1 + k2
    Pos = gammaln(np.sum(k1, axis=0) + 1) + gammaln(np.sum(k2, axis=0) + 1) + np.tile(gammaln(np.sum(alpha)), (1, Ks)) + np.sum(gammaln(np.tile(alpha, (1, Ks)) + k), axis=0)
    Neg = np.sum(gammaln(k1 + 1), axis=0) + np.sum(gammaln(k2 + 1), axis=0) + gammaln(np.sum(np.tile(alpha, (1, Ks)) + k, axis=0)) + np.tile(np.sum(gammaln(alpha)), (1, Ks))
    log = Pos - Neg
    return log



"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from collections import defaultdict
from scipy.special import logsumexp
from scipy.sparse import csr_matrix
from scipy.stats import nbinom
import h5py
import itertools
import multiprocessing
import numpy as np
import pandas as pd
import time

from omniCLIP import viterbi
from omniCLIP.data_parsing import LoadReads
from omniCLIP.omni_stat import diag_event_model
from omniCLIP.omni_stat import emission_prob
from omniCLIP.omni_stat import trans


def GetModelIx(Sequences, Type='all', snps_thresh=0.4, snps_min_cov=10,
               Background=None):
    """Return the positions for at which the emission should be computed."""
    if Type == 'all':
        Ix = np.sum(StackData(Sequences, add='all'), axis=0) > 0
    elif Type == 'no_snps_all':
        if Background is not None:
            Bck_var = np.sum(StackData(Background, add='only_var'), axis=0)
            Bck = np.sum(StackData(Background, add='only_cov'), axis=0)
            Bck_ratio = np.zeros_like(Bck_var)
            ix = Bck > 0
            Bck_ratio[ix] = np.float64(Bck_var[ix]) / np.float64(Bck[ix])
            # Positions that have a minimal coverage and have a snp ratio above
            # the threshold are not considered
            Ix_bg = (Bck > snps_min_cov) * (Bck_ratio > snps_thresh) > 0
            Ix_fg = np.sum(StackData(Sequences, add='all'), axis=0) > 0
            Ix = (Ix_fg > 0) * (Ix_bg == 0) > 0
    elif Type == 'no_snps_conv':
        if Background is not None:
            Bck_var = np.sum(StackData(Background, add='only_var'), axis=0)
            Bck = np.sum(StackData(Background, add='only_cov'), axis=0)
            Bck_ratio = np.zeros_like(Bck_var)
            ix = Bck > 0
            Bck_ratio[ix] = np.float64(Bck_var[ix]) / np.float64(Bck[ix])
            # Positions that have a minimal coverage and have a snp ratio above
            # the threshold are not considered
            Ix_bg = (Bck > snps_min_cov) * (Bck_ratio > snps_thresh) > 0
            Ix_fg = np.sum(StackData(Sequences, add='nocov'), axis=0) > 0
            Ix = (Ix_fg > 0) * (Ix_bg == 0) > 0
    elif Type == 'Conv':
        Ix = np.sum(StackData(Sequences, add='nocov'), axis=0) > 0

    else:
        Ix = np.sum(StackData(Sequences, add='nocov'), axis=0) > 0
    return Ix


def StackData(Sequences, add='all', use_strand='True'):
    """Stack the data for a gene."""
    Track_strand_map = [18, 17, 16, 15, 19, 13, 12, 11, 10, 14, 8, 7, 6, 5, 9,
                        3, 2, 1, 0, 4]

    repls = list(Sequences['Coverage'].keys())
    nr_of_repl = len(repls)
    gene_len = Sequences['Coverage']['0'].shape[1]
    if (add == 'all') or (add == 'nocov'):
        CurrStack = np.zeros((nr_of_repl, gene_len))
        for rep in repls:
            CurrStack[int(rep), :] = Sequences['SummedCoverage'][rep]

    elif add == 'only_cov':
        # Check if the cariants are substracted from the coverage:
        if 'Variants' in Sequences:
            CurrStack = np.zeros((nr_of_repl, gene_len))
            for rep in repls:
                CurrStack[int(rep), :] = Sequences['SummedCoverage'][rep]
        else:
            CurrStack = np.zeros((nr_of_repl, gene_len))
            for rep in repls:
                CurrStack[int(rep), :] += Sequences['Coverage'][rep].sum(axis=0)
    elif add == 'only_var':
        CurrStack = np.zeros((nr_of_repl, gene_len))
        for rep in repls:
            CurrStack[int(rep), :] += Sequences['Variants'][rep].sum(axis=0)

    elif add == 'only_var_summed':
        CurrStack = np.zeros(
            (Sequences['Variants'][repls[0]].shape[0], gene_len))
        for rep in repls:
            if use_strand and (not (Sequences['strand'] == 1)):
                CurrStack += Sequences['Variants'][rep][Track_strand_map, :]
            else:
                CurrStack += Sequences['Variants'][rep]

    else:
        # First compute dimensions of return array:
        nr_rows = 0
        for rep in list(Sequences['Variants'].keys()):
            nr_cols = Sequences['Coverage'][rep].shape[1]
            nr_rows += (Sequences['Variants'][rep].shape[0]
                        + Sequences['Read-ends'][rep].shape[0]
                        + Sequences['Coverage'][rep].shape[0])

        CurrStack = np.zeros((nr_rows, nr_cols))
        row_counter = 0
        for rep in list(Sequences['Variants'].keys()):
            if use_strand and (not (Sequences['strand'] == 1)):
                CurrStack[row_counter:(row_counter + Sequences['Variants'][rep].shape[0]), :] = Sequences['Variants'][rep][Track_strand_map, :]
            else:
                CurrStack[row_counter:(row_counter + Sequences['Variants'][rep].shape[0]), :] = Sequences['Variants'][rep]
            row_counter += Sequences['Variants'][rep].shape[0]

            CurrStack[row_counter:(row_counter + Sequences['Read-ends'][rep].shape[0]), :] = Sequences['Read-ends'][rep]
            row_counter += Sequences['Read-ends'][rep].shape[0]

            CurrStack[row_counter:(row_counter + Sequences['Coverage'][rep].shape[0]), :] = Sequences['Coverage'][rep]
            row_counter += Sequences['Coverage'][rep].shape[0]

    CurrStack[CurrStack < 0] = 0
    return CurrStack


def PreloadSequencesForGene(Sequences, gene):
    """Stacks the data for a gene."""
    Sequences_per_gene = {}
    for key in Sequences[gene]:
        if isinstance(Sequences[gene][key], h5py.Dataset):
            Sequences_per_gene[key] = Sequences[gene][key][()]
        else:
            Sequences_per_gene[key] = {}
            for rep in Sequences[gene][key]:
                if key == 'Variants':
                    # Convert the Variants to array again
                    Sequences_per_gene[key][rep] = csr_matrix(
                        (Sequences[gene]['Variants'][rep]['data'][:],
                         Sequences[gene]['Variants'][rep]['indices'][:],
                         Sequences[gene]['Variants'][rep]['indptr'][:]),
                        shape=Sequences[gene]['Variants'][rep]['shape'][:]).toarray()
                else:
                    Sequences_per_gene[key][rep] = Sequences[gene][key][rep][()]

    return Sequences_per_gene


def GetSuffStat(Paths, NrOfStates, Type, ResetNotUsedStates=True,
                EmissionParameters=None, verbosity=1):
    """Compute for each CurrPath state a set of sufficient statistics."""
    # Initialize the sufficent statistics variable
    print("Getting suffcient statistic")
    t = time.time()
    SuffStat = {}
    for CurrState in range(NrOfStates):
        SuffStat[CurrState] = defaultdict(int)

    Sequences = h5py.File(EmissionParameters['dat_file_clip'], 'r')
    Background = h5py.File(EmissionParameters['dat_file_bg'], 'r')

    # Fill the sufficent statistics variable
    for gene in list(Sequences.keys()):
        rep = list(Sequences[gene]['Coverage'].keys())[0]
        CurrGenePath = Paths[gene]

        # Stack the matrices together and convert to dense matrix
        Sequences_per_gene = PreloadSequencesForGene(Sequences, gene)
        Background_per_gene = PreloadSequencesForGene(Background, gene)
        if Type == 'Conv':
            CurrStack = StackData(Sequences_per_gene, add='variants')
        else:
            CurrStack = StackData(Sequences_per_gene, add='all')

        if EmissionParameters['filter_snps']:
            if Type == 'Conv':
                Ix = GetModelIx(
                    Sequences_per_gene, Type='no_snps_conv',
                    snps_thresh=EmissionParameters['snps_thresh'],
                    snps_min_cov=EmissionParameters['snps_min_cov'],
                    Background=Background_per_gene)
            else:
                Ix = GetModelIx(Sequences_per_gene, Type)
        else:
            Ix = GetModelIx(Sequences_per_gene, Type)

        NonZero = np.sum(CurrStack, axis=0) > 0

        # Determine the nonzeros elements
        for CurrState in range(NrOfStates):
            if EmissionParameters['mask_ovrlp']:
                CurrIx = Ix * (Sequences_per_gene['mask'][rep][0, :] == 0) * NonZero * (CurrGenePath == CurrState) > 0
            else:
                CurrIx = Ix * NonZero * (CurrGenePath == CurrState) > 0

            data = CurrStack[:, CurrIx].T
            ncols = data.shape[1]
            dtype = data.T.dtype.descr * ncols
            struct = data.view(dtype)

            vals, val_counts = np.unique(struct, return_counts=True)

            # Save the tuples and how many times they have been seen so far.
            for curr_val, curr_count in zip(vals, val_counts):
                SuffStat[CurrState][tuple(curr_val)] += curr_count

            # Treat the 0 tuple seperately for speed improvment
            if len(Ix) == 0:
                continue

            NullIx = (NonZero == 0) * (CurrGenePath == CurrState) > 0
            if np.sum(NullIx) == 0:
                continue
            NullCount = np.sum(NullIx)
            if NullCount > 0:
                NullTuple = np.zeros_like(CurrStack[:, 0])
                NullTuple = tuple(NullTuple.T)
                SuffStat[CurrState][NullTuple] += NullCount

        del CurrStack, NonZero, CurrGenePath, Ix
    if verbosity > 0:
        print('Done: Elapsed time: ' + str(time.time() - t))

    return SuffStat


def GetSuffStatBck(Paths, NrOfStates, Type, ResetNotUsedStates=True,
                   EmissionParameters=None, verbosity=1):
    """Computes for each CurrPath state a set of sufficient statistics."""
    # Initialize the sufficent statistics variable
    print("Getting suffcient statistic")
    t = time.time()
    SuffStatBck = {}

    fg_state, bg_state = emission_prob.get_fg_and_bck_state(
        EmissionParameters, final_pred=True)

    SuffStatBck[fg_state] = defaultdict(int)

    Sequences = h5py.File(EmissionParameters['dat_file_clip'], 'r')
    Background = h5py.File(EmissionParameters['dat_file_bg'], 'r')

    # Fill the sufficient statistics variable
    for gene in list(Sequences.keys()):
        rep = list(Background[gene]['Coverage'].keys())[0]
        CurrGenePath = Paths[gene]

        # Stack the matrices together and convert to dense matrix
        Background_per_gene = PreloadSequencesForGene(Background, gene)
        Sequences_per_gene = PreloadSequencesForGene(Sequences, gene)
        if Type == 'Conv':
            CurrStack = StackData(Background_per_gene, add='variants')
        else:
            CurrStack = StackData(Background_per_gene, add='all')

        if EmissionParameters['filter_snps']:
            if Type == 'Conv':
                Ix = GetModelIx(
                    Background_per_gene, Type='no_snps_conv',
                    snps_thresh=EmissionParameters['snps_thresh'],
                    snps_min_cov=EmissionParameters['snps_min_cov'],
                    Background=Background_per_gene)
            else:
                Ix = GetModelIx(Background_per_gene, Type)
        else:
            Ix = GetModelIx(Background_per_gene, Type)

        NonZero = np.sum(CurrStack, axis=0) > 0

        # Determine the nonzeros elements
        CurrState = fg_state

        CurrIx = Ix * NonZero > 0
        if EmissionParameters['mask_ovrlp']:
            CurrIx = (Ix * NonZero * (CurrGenePath == CurrState)
                      * (Sequences_per_gene['mask'][rep][0, :] == 0)) > 0
        else:
            CurrIx = Ix * NonZero * (CurrGenePath == CurrState) > 0

        data = CurrStack[:, CurrIx].T
        ncols = data.shape[1]
        dtype = data.T.dtype.descr * ncols
        struct = data.view(dtype)

        vals, val_counts = np.unique(struct, return_counts=True)

        # Save the tuples and how many times they have been seen so far.
        for curr_val, curr_count in zip(vals, val_counts):
            SuffStatBck[CurrState][tuple(curr_val)] += curr_count

        # Treat the 0 tuple seperately for speed improvment
        if len(Ix) == 0:
            continue
        NullIx = (NonZero == 0) * (CurrGenePath == CurrState) > 0
        if np.sum(NullIx) == 0:
            continue
        NullCount = np.sum(NullIx)
        if NullCount > 0:
            NullTuple = np.zeros_like(CurrStack[:, 0])
            NullTuple = tuple(NullTuple.T)
            SuffStatBck[CurrState][NullTuple] += NullCount

        del CurrStack, NonZero, CurrGenePath, Ix

    print('Done: Elapsed time: ' + str(time.time() - t))

    return SuffStatBck


def ConvertSuffStatToArrays(SuffStat):
    """Convert the sufficient statistics into a list of arrays."""
    # Initialise the return values
    Counts = {}
    NrOfCounts = {}
    for State in list(SuffStat.keys()):
        if len(list(SuffStat[State].keys())) == 0:
            print("empyt suffcient keys ")
            Counts[State] = np.array([])
            NrOfCounts[State] = np.array([])
        else:
            Counts[State] = np.array(
                [np.array(key) for key in list(SuffStat[State].keys())]).T
            NrOfCounts[State] = np.tile(
                np.array(list(SuffStat[State].values())), (1, 1))

    return Counts, NrOfCounts


def repl_track_nr(ex_list, offset):
    """Compute replicate list offset.

    Compute for a list of tracks in one replicate additionaly the list for the
    second replicate.
    """
    return ex_list + list(np.array(ex_list) + offset)


def GeneratePred(Paths, Sequences, Background, IterParameters, GeneAnnotation,
                 OutFile, fg_state=1, noise_state=0, pv_cutoff=0.05,
                 verbosity=1):
    """Write the predictions."""
    TransitionParameters = IterParameters[1]
    EmissionParameters = IterParameters[0]

    merge_neighbouring_sites = False
    minimal_site_length = 1

    # Predict the sites
    print('Score peaks')
    LoadReads.close_data_handles(handles=[Sequences, Background])
    Sequences = h5py.File(EmissionParameters['dat_file_clip'], 'r')
    Background = h5py.File(EmissionParameters['dat_file_bg'], 'r')

    ScoredSites = GetSites(
        Paths, Sequences, Background, EmissionParameters,
        TransitionParameters, 'nonhomo', fg_state, merge_neighbouring_sites,
        minimal_site_length, seq_file=EmissionParameters['dat_file_clip'],
        bck_file=EmissionParameters['dat_file_bg'])

    Sequences = h5py.File(EmissionParameters['dat_file_clip'], 'r')
    Background = h5py.File(EmissionParameters['dat_file_bg'], 'r')

    print('Write peaks')
    # Write the results
    WriteResults(Sequences, Background, ScoredSites, OutFile, GeneAnnotation)

    generate_bed(OutFile, pv_cutoff=0.05)
    return


def generate_bed(file, pv_cutoff=0.05):
    df = pd.read_table(file, low_memory=False)

    # Determine the cutoff for the p-vales aften Bonferroni
    cutoff = np.log(0.05 / df.shape[0])
    df = df[df['pv'] < cutoff]

    # Make the names unique
    df['Gene'] = df['Gene'] + ['_' + str(i) for i in range(df.shape[0])]

    # Transform the Score to be betweeen 0 and 1000
    df['SiteScore'] /= np.max(df['SiteScore']) * 0.001

    # Compute thick start and end
    df['ThickStart'] = 0
    df['ThickStop'] = 0

    df.loc[df['Strand'] == 1, 'ThickStart'] = df['Start'] + df['max_pos']
    df.loc[df['Strand'] == -1, 'ThickStart'] = df['Stop'] - df['max_pos']

    df.loc[df['Strand'] == 1, 'ThickStop'] = df['Start'] + df['max_pos'] + 1
    df.loc[df['Strand'] == -1, 'ThickStop'] = df['Stop'] - df['max_pos'] + 1

    # Rename the strands
    df.loc[df['Strand'] == 1, 'Strand'] = '+'
    df.loc[df['Strand'] == -1, 'Strand'] = '-'

    # Keep only the columns that are relevant for the bed-file
    df = df[
        ['ChrName', 'Start', 'Stop', 'Gene',
         'SiteScore', 'Strand', 'ThickStart', 'ThickStop']
    ].sort_values('SiteScore', ascending=False)

    # Write the output
    df.to_csv(file.replace('.txt', '.bed'), sep='\t', header=False, index=False)
    return


def GetSites(Paths, Sequences, Background, EmissionParameters,
             TransitionParameters, TransitionTypeFirst, fg_state,
             merge_neighbouring_sites, minimal_site_length, seq_file='',
             bck_file=''):
    """Get the predicted sites."""
    # Extract the paths
    Sites = convert_paths_to_sites(
        Paths, fg_state,
        merge_neighbouring_sites, minimal_site_length)

    nr_of_genes = len(list(Sequences.keys()))
    gene_nr_dict = {}
    for i, curr_gene in enumerate(Sequences.keys()):
        gene_nr_dict[curr_gene] = i

    sites_keys = [key for key in list(Sites.keys()) if len(Sites[key]) > 0]
    f = lambda key, Sites=Sites, nr_of_genes=nr_of_genes, gene_nr_dict=gene_nr_dict, seq_file=seq_file, bck_file=bck_file, EmissionParameters=EmissionParameters, TransitionParameters=TransitionParameters, TransitionTypeFirst=TransitionTypeFirst, fg_state=fg_state, merge_neighbouring_sites=merge_neighbouring_sites, minimal_site_length=minimal_site_length: (Sites[key], key, nr_of_genes, gene_nr_dict[key], seq_file, bck_file, EmissionParameters, TransitionParameters, TransitionTypeFirst, fg_state, merge_neighbouring_sites, minimal_site_length)

    data = map(f, sites_keys)

    LoadReads.close_data_handles(handles=[Sequences, Background])
    if EmissionParameters['nb_proc'] == 1:
        ScoredSites = dict([GetSitesForGene(curr_slice) for curr_slice in data])
    else:
        pool = multiprocessing.get_context("spawn").Pool(
            EmissionParameters['nb_proc'], maxtasksperchild=10)
        results = pool.imap(GetSitesForGene, data, chunksize=1)
        pool.close()
        pool.join()
        ScoredSites = dict([res for res in results])
    for key in list(ScoredSites.keys()):
        if len(ScoredSites[key]) == 0:
            del ScoredSites[key]

    return ScoredSites


def GetSitesForGene(data):
    """Determine the score of the sites for each gene."""
    # Computing the probabilities for the current gene
    (Sites, gene, nr_of_genes, gene_nr, seq_file, bck_file,
     EmissionParameters, TransitionParameters, TransitionTypeFirst,
     fg_state, merge_neighbouring_sites, minimal_site_length) = data

    # Turn the Sequence and Bacground objects into dictionaries again such that
    # the subsequent methods for using these do not need to be modified
    if len(Sites) == 0:
        return gene, []

    NrOfStates = EmissionParameters['NrOfStates']
    TransitionType = EmissionParameters['TransitionType']

    Sites = dict([(gene, Sites)])

    Sequences = h5py.File(EmissionParameters['dat_file_clip'], 'r')
    Background = h5py.File(EmissionParameters['dat_file_bg'], 'r')

    Sequences_per_gene = PreloadSequencesForGene(Sequences, gene)
    Background_per_gene = PreloadSequencesForGene(Background, gene)

    Ix = GetModelIx(Sequences_per_gene, Type='all')

    if np.sum(Ix) == 0:
        return gene, []

    if EmissionParameters['filter_snps']:
        Ix = GetModelIx(Sequences_per_gene, Type='no_snps_conv',
                        snps_thresh=EmissionParameters['snps_thresh'],
                        snps_min_cov=EmissionParameters['snps_min_cov'],
                        Background=Background_per_gene)
    else:
        Ix = GetModelIx(Sequences_per_gene, Type='Conv')

    # Only compute the emission probability for regions where a site is
    ix_sites = np.zeros_like(Ix)
    ix_sites_len = Ix.shape[0]
    for currsite in Sites[gene]:
        ix_sites[max(0, currsite[0] - 1): min(ix_sites_len, currsite[1] + 1)] = 1
    ix_sites = ix_sites == 1

    # 2) Compute the probabilities for both states
    EmmisionProbGene = np.log(
        np.ones((NrOfStates, Ix.shape[0])) * (1 / np.float64(NrOfStates)))
    CurrStackSum = StackData(Sequences_per_gene)
    CurrStackVar = StackData(Sequences_per_gene, add='no')
    CurrStackSumBck = StackData(Background_per_gene, add='only_cov')

    CurrStackVarSumm = StackData(Sequences_per_gene, add='only_var_summed')
    EmmisionProbGeneDir = np.zeros_like(EmmisionProbGene)

    if EmissionParameters['glm_weight'] < 0.0:
        weight1 = 1.0
        weight2 = 1.0
    elif EmissionParameters['glm_weight'] == 0.0:
        weight1 = 0.0000001
        weight2 = 1.0 - weight1
    elif EmissionParameters['glm_weight'] == 1.0:
        weight1 = 0.9999999
        weight2 = 1.0 - weight1
    else:
        weight1 = EmissionParameters['glm_weight']
        weight2 = (1.0 - EmissionParameters['glm_weight'])

    for State in range(NrOfStates):
        EmmisionProbGene[State, ix_sites] = np.log(weight1) + emission_prob.predict_expression_log_likelihood_for_gene(CurrStackSum[:, ix_sites], State, nr_of_genes, gene_nr, EmissionParameters)
        if EmissionParameters['bg_type'] == 'Coverage':
            EmmisionProbGene[State, ix_sites] += np.log(weight1) + emission_prob.predict_expression_log_likelihood_for_gene(CurrStackSumBck[:, ix_sites], State, nr_of_genes, gene_nr, EmissionParameters, 'bg')
        if EmissionParameters['bg_type'] == 'Coverage_bck':
            EmmisionProbGene[State, ix_sites] += np.log(weight1) + emission_prob.predict_expression_log_likelihood_for_gene(CurrStackSumBck[:, ix_sites], State, nr_of_genes, gene_nr, EmissionParameters, 'bg')
        EmmisionProbGeneDir[State, Ix] = np.log(weight2) + diag_event_model.pred_log_lik(CurrStackVar[:, Ix], State, EmissionParameters)
        EmmisionProbGene[State, Ix] += np.log(weight2) + EmmisionProbGeneDir[State, Ix]

    if TransitionType == 'unif_bck' or TransitionType == 'binary_bck':
        CountsSeq = StackData(Sequences_per_gene, add='all')
        CountsBck = StackData(Background_per_gene, add='only_cov')
        Counts = np.vstack((CountsSeq, CountsBck))
    else:
        Counts = StackData(Sequences_per_gene, add='all')

    Score = EmmisionProbGene
    CurrStack = CurrStackVar
    # Compute the scores when staying in the same state
    # RowIx = list(range(16)) + list(range(17, 38)) + list(range(39,44))
    strand = Sequences_per_gene['strand']

    # Get the coverages for the froeground and background
    CountsSeq = StackData(Sequences_per_gene, add='only_cov')
    CountsBck = StackData(Background_per_gene, add='only_cov')

    if strand == 0:
        strand = -1
    # Since we the transition probabilty is the same for all States we do not
    # need to compute it for the bayes factor this list contains the returned
    # sites
    sites = []
    for currsite in Sites[gene]:
        (mean_mat_fg, var_mat_fg, mean_mat_bg, var_mat_bg, counts_fg, counts_bg) = ComputeStatsForSite(CountsSeq, CountsBck, currsite, fg_state, nr_of_genes, gene_nr, EmissionParameters)

        site = {}
        site['Start'] = currsite[0]
        site['Stop'] = currsite[1]
        site['Strand'] = strand
        site['SiteScore'] = EvaluateSite(Score, currsite, fg_state)
        site['Coverage'] = np.sum(np.sum(Counts[:, site['Start']:site['Stop']], axis=0))
        site['Variants'] = np.sum(CurrStackVarSumm[:, site['Start']:site['Stop']], axis=1)
        site['mean_mat_fg'] = mean_mat_fg
        site['var_mat_fg'] = var_mat_fg
        site['mean_mat_bg'] = mean_mat_bg
        site['var_mat_bg'] = var_mat_bg
        site['counts_fg'] = counts_fg
        site['counts_bg'] = counts_bg

        p = mean_mat_fg / var_mat_fg
        n = (mean_mat_fg ** 2) / (var_mat_fg - mean_mat_fg)
        site['pv'] = nbinom.logsf(counts_fg, n, p)
        site['max_pos'] = get_max_position(Score, currsite, fg_state, strand)
        site['dir_score'] = EvaluateSite(EmmisionProbGeneDir, currsite, fg_state)
        if site['SiteScore'] < 0.0:
            continue
        sites.append(site)

    LoadReads.close_data_handles(handles=[Sequences, Background])

    return gene, sites


def convert_paths_to_sites(Paths, fg_state, merge_neighbouring_sites,
                           minimal_site_length):
    """Compute the site predictions using the paths."""
    sites = defaultdict(list)

    # Iterate over the paths
    for gene in Paths:
        curr_path = Paths[gene] == fg_state
        Starts = np.where(np.concatenate(([curr_path[0]], curr_path[:-1] != curr_path[1:], [curr_path[-1]])))[0][::2]
        Stops = np.where(np.concatenate(([curr_path[0]], curr_path[:-1] != curr_path[1:], [curr_path[-1]])))[0][1::2]
        nr_sites = Starts.shape[0]
        sites[gene] = [[Starts[i], Stops[i]] for i in range(nr_sites) if (Stops[i] - Starts[i] >= minimal_site_length)]

    return sites


def ComputeStatsForSite(CountsSeq, CountsBck, Site, fg_state, nr_of_genes,
                        gene_nr, EmissionParameters):
    """Get the score for a Site."""
    Start = Site[0]
    Stop = Site[1]

    mean_mat_fg, var_mat_fg = emission_prob.get_expected_mean_and_var(
        CountsSeq[:, Start:(Stop + 1)], fg_state, nr_of_genes,
        gene_nr, EmissionParameters, curr_type='fg')
    mean_mat_bg, var_mat_bg = emission_prob.get_expected_mean_and_var(
        CountsBck[:, Start:(Stop + 1)], fg_state, nr_of_genes,
        gene_nr, EmissionParameters, curr_type='bg')

    mean_mat_fg = np.sum(np.sum(mean_mat_fg, axis=0))
    var_mat_fg = np.sum(np.sum(var_mat_fg, axis=0))
    mean_mat_bg = np.sum(np.sum(mean_mat_bg, axis=0))
    var_mat_bg = np.sum(np.sum(var_mat_bg, axis=0))
    counts_fg = np.sum(np.sum(CountsSeq[:, Start:(Stop + 1)], axis=0))
    counts_bg = np.sum(np.sum(CountsBck[:, Start:(Stop + 1)], axis=0))

    return mean_mat_fg, var_mat_fg, mean_mat_bg, var_mat_bg, counts_fg, counts_bg


def get_max_position(Score, Site, fg_state, strand):
    """Get the site where the score is maximal."""
    Start = Site[0]
    Stop = Site[1]
    ix_bg = list(range(Score.shape[0]))
    ix_bg.remove(fg_state)
    FGScore = Score[fg_state, Start:(Stop + 1)]
    AltScore = Score[ix_bg, Start:(Stop + 1)]

    norm = logsumexp(AltScore, axis=0)

    ix_ok = np.isinf(norm) + np.isnan(norm)
    if np.sum(ix_ok) < norm.shape[0]:
        SiteScore = FGScore[ix_ok == 0] - norm[ix_ok == 0]
    else:
        print('Score problematic')
        SiteScore = FGScore

    max_pos = np.int64(np.round(np.mean(np.where(SiteScore == np.max(SiteScore))[0])))
    if strand == -1:
        pos = Stop - Start - max_pos
    else:
        pos = max_pos

    return pos


def EvaluateSite(Score, Site, fg_state):
    """Get the score for a Site."""
    Start = Site[0]
    Stop = Site[1]
    ix_bg = list(range(Score.shape[0]))
    ix_bg.remove(fg_state)
    FGScore = np.sum(Score[fg_state, Start:(Stop + 1)])
    AltScore = np.sum(Score[ix_bg, Start:(Stop + 1)], axis=1)
    norm = logsumexp(AltScore)
    if not (np.isinf(norm) or np.isnan(norm)):
        SiteScore = FGScore - norm
    else:
        print('Score problematic')
        SiteScore = FGScore

    return SiteScore


def WriteResults(Sequences, Background, ScoredSites, OutFile, GeneAnnotation):
    """Write the sites into a result file."""
    # Get the gene annotation
    Iter = GeneAnnotation.features_of_type('gene')
    Genes = []
    for gene in Iter:
        Genes.append(gene)

    # Print the results
    fid = open(OutFile, 'w')
    VariantList = [Nucl_1 + '-' + Nucl_2
                   for Nucl_1 in ['A', 'C', 'G', 'T']
                   for Nucl_2 in ['A', 'C', 'G', 'T', 'D']]
    Header = '\t'.join(
        ['Gene', 'ChrName', 'Start', 'Stop', 'Strand', 'SiteScore', 'Coverage']
        + VariantList
        + ['mean_mat_fg', 'var_mat_fg', 'mean_mat_bg', 'var_mat_bg',
           'counts_fg', 'counts_bg', 'pv', 'max_pos', 'dir_score']) + '\n'
    fid.write(Header)
    for gene in Genes:
        gene_name = gene.id
        if gene_name not in ScoredSites:
            continue
        # Transform the Coordinates
        for site in ScoredSites[gene_name]:
            # Process the current site
            CurrLine = '\t'.join(
                [gene_name, gene.chrom,
                 str(GetGenomicCoord(gene, site['Start'])),
                 str(GetGenomicCoord(gene, site['Stop'])),
                 str(site['Strand']), str(site['SiteScore']),
                 str(site['Coverage'])]
                + [str(int(site['Variants'][i]))
                   for i in range(len(site['Variants']))]
                + [str(site['mean_mat_fg']), str(site['var_mat_fg']),
                   str(site['mean_mat_bg']), str(site['var_mat_bg']),
                   str(site['counts_fg']), str(site['counts_bg']),
                   str(site['pv']), str(site['max_pos']),
                   str(site['dir_score'])]
            ) + '\n'

            fid.write(CurrLine)
    fid.close()

    return


def GetGenomicCoord(gene, Coord):
    """Return formatted gene coordinates."""
    return gene.start + Coord


def estimate_library_size(Sequences):
    """Estimate the library size of all samples."""
    lib_size_dict = defaultdict(list)
    lib_size_red = defaultdict(int)
    # Get the gene expressions
    for gene in list(Sequences.keys()):
        for key in list(Sequences[gene]['Coverage'].keys()):
            lib_size_dict[key].append(
                Sequences[gene]['Coverage'][key][()].sum())

    # Compute the (weighted) median of non zero genes
    for key in lib_size_dict:
        exprs = np.array([g for g in lib_size_dict[key] if g > 0])

        lib_size_red[key] = np.median(exprs)

    return lib_size_red


def ParallelGetMostLikelyPath(
        MostLikelyPaths, Sequences, Background, EmissionParameters,
        TransitionParameters, TransitionTypeFirst, RandomNoise=False,
        chunksize=1, verbosity=1):
    """Compute the most likely path.

    There are two options, 'homo' and 'nonhomo' for TransitionType. This
    specifies whether the transition probabilities should be homogenous or
    non-homogenous.
    """

    for gene in list(MostLikelyPaths.keys()):
        del MostLikelyPaths[gene]
    nr_of_genes = len(list(Sequences.keys()))
    gene_nr_dict = {}
    for i, curr_gene in enumerate(Sequences.keys()):
        gene_nr_dict[curr_gene] = i

    print("Computing most likely path")
    t = time.time()

    data = zip(list(Sequences.keys()), itertools.repeat(nr_of_genes),
               list(gene_nr_dict.values()),
               itertools.repeat(EmissionParameters),
               itertools.repeat(TransitionParameters),
               itertools.repeat(TransitionTypeFirst),
               itertools.repeat(RandomNoise))

    LoadReads.close_data_handles(handles=[Sequences, Background])

    if EmissionParameters['nb_proc'] == 1:
        results = [ParallelGetMostLikelyPathForGene(curr_slice)
                   for curr_slice in data]
    else:
        print("Spawning processes")
        pool = multiprocessing.get_context("spawn").Pool(
            EmissionParameters['nb_proc'], maxtasksperchild=5)
        results = pool.imap(ParallelGetMostLikelyPathForGene, data, chunksize)
        pool.close()
        pool.join()
        print("Collecting results")
        results = [res for res in results]

    MostLikelyPaths = dict(zip(
        [result[0] for result in results], [result[1] for result in results]))

    # Compute the logliklihood of the gene
    LogLikelihood = sum([result[2] for result in results])
    del results

    if verbosity > 0:
        print('\nDone: Elapsed time: ' + str(time.time() - t))

    return MostLikelyPaths, LogLikelihood


def ParallelGetMostLikelyPathForGene(data):
    """Compute the most likely path for a gene."""
    (gene, nr_of_genes, gene_nr, EmissionParameters,
     TransitionParameters, TransitionTypeFirst, RandomNoise) = data

    # Turn the Sequence and Bacground objects into dictionaries again such that
    # the subsequent methods for using these do not need to be modified
    Sequences = h5py.File(EmissionParameters['dat_file_clip'], 'r')
    Background = h5py.File(EmissionParameters['dat_file_bg'], 'r')

    # Parse the parameters
    alpha = EmissionParameters['Diag_event_params']
    PriorMatrix = EmissionParameters['PriorMatrix']
    NrOfStates = EmissionParameters['NrOfStates']

    TransitionType = EmissionParameters['TransitionType']

    fg_state, bg_state = emission_prob.get_fg_and_bck_state(
        EmissionParameters, final_pred=True)
    # Score the state sequences
    # 1) Determine the positions where an observation is possible

    Sequences_per_gene = PreloadSequencesForGene(Sequences, gene)
    Background_per_gene = PreloadSequencesForGene(Background, gene)

    Ix = GetModelIx(Sequences_per_gene, Type='all')

    if np.sum(Ix) == 0:
        CurrPath = 2 * np.ones((0, Ix.shape[0]), dtype=np.int)
        return [gene, CurrPath, 0]

    if EmissionParameters['filter_snps']:
        Ix = GetModelIx(Sequences_per_gene, Type='no_snps_conv',
                        snps_thresh=EmissionParameters['snps_thresh'],
                        snps_min_cov=EmissionParameters['snps_min_cov'],
                        Background=Background_per_gene)
    else:
        Ix = GetModelIx(Sequences_per_gene)

    # 2) Compute the probabilities for both states
    EmmisionProbGene = (np.ones((NrOfStates, Ix.shape[0]))
                        * (1 / np.float64(NrOfStates)))

    CurrStackSum = StackData(Sequences_per_gene)
    CurrStackVar = StackData(Sequences_per_gene, add='no')
    CurrStackSumBck = StackData(Background_per_gene, add='only_cov')

    if EmissionParameters['glm_weight'] < 0.0:
        weight1 = 1.0
        weight2 = 1.0
    elif EmissionParameters['glm_weight'] == 0.0:
        weight1 = 0.0000001
        weight2 = 1.0 - weight1
    elif EmissionParameters['glm_weight'] == 1.0:
        weight1 = 0.9999999
        weight2 = 1.0 - weight1
    else:
        weight1 = EmissionParameters['glm_weight']
        weight2 = (1.0 - EmissionParameters['glm_weight'])

    for State in range(NrOfStates):
        if not EmissionParameters['ign_GLM']:
            if isinstance(EmissionParameters['ExpressionParameters'][0], np.ndarray):
                EmmisionProbGene[State, :] = np.log(weight1) + emission_prob.predict_expression_log_likelihood_for_gene(CurrStackSum, State, nr_of_genes, gene_nr, EmissionParameters)
                if EmissionParameters['bg_type'] == 'Coverage':
                    EmmisionProbGene[State, :] += np.log(weight1) + emission_prob.predict_expression_log_likelihood_for_gene(CurrStackSumBck, State, nr_of_genes, gene_nr, EmissionParameters, 'bg')
                if EmissionParameters['bg_type'] == 'Coverage_bck':
                    EmmisionProbGene[State, :] += np.log(weight1) + emission_prob.predict_expression_log_likelihood_for_gene(CurrStackSumBck, State, nr_of_genes, gene_nr, EmissionParameters, 'bg')
        if not EmissionParameters['ign_diag']:
            EmmisionProbGene[State, Ix] += np.log(weight2) + diag_event_model.pred_log_lik(CurrStackVar[:, Ix], State, EmissionParameters)
        if State == fg_state:
            if EmissionParameters['LastIter']:
                EmmisionProbGene[State, :] -= EmissionParameters['fg_pen']
    if RandomNoise:
        EmmisionProbGene = np.logaddexp(
            EmmisionProbGene, np.random.uniform(
                np.min(EmmisionProbGene[np.isfinite(EmmisionProbGene)]) - 4,
                np.min(EmmisionProbGene[np.isfinite(EmmisionProbGene)]) - 0.1,
                EmmisionProbGene.shape))  # Add some random noise

    # Get the transition probabilities
    if TransitionTypeFirst == 'nonhomo':
        if TransitionType == 'unif_bck' or TransitionType == 'binary_bck':
            CountsSeq = StackData(Sequences_per_gene, add='all')
            CountsBck = StackData(Background_per_gene, add='only_cov')
            Counts = np.vstack((CountsSeq, CountsBck))
        else:
            Counts = StackData(Sequences_per_gene, add='all')
        TransistionProbabilities = np.float64(trans.PredictTransistions(
            Counts, TransitionParameters, NrOfStates, TransitionType))
    else:
        TransistionProbabilities = np.float64(np.tile(
            np.log(TransitionParameters[0]), (EmmisionProbGene.shape[1],1,1)).T)

    CurrPath, Currloglik = viterbi.viterbi(
        np.float64(EmmisionProbGene), TransistionProbabilities,
        np.float64(np.log(PriorMatrix)))
    CurrPath = np.int8(CurrPath)

    del (TransistionProbabilities, EmmisionProbGene, CurrStackSum,
         CurrStackVar, CurrStackSumBck, Ix)

    LoadReads.close_data_handles(handles=[Sequences, Background])

    return [gene, CurrPath, Currloglik]


def subsample_suff_stat(Counts, NrOfCounts, subsample_size=250000):
    """Create a subsample of the counts."""
    # Iterate over the keys
    for key in Counts:
        # Determine the new sample
        if NrOfCounts[key].shape[0] == 0:
            # Try to rescue parameter fitting by adding some data from other
            # regions for fitting.
            for temp_key in Counts:
                if NrOfCounts[temp_key].shape[0] > 10:
                    NrOfCounts[key] = np.ones((1, 10))
                    ix = np.random.randint(
                        NrOfCounts[temp_key].shape[0], size=10)
                    Counts[key] = Counts[temp_key][:, ix]
                    break

        new_counts = np.random.multinomial(
            min(subsample_size, np.sum(NrOfCounts[key][0, :])),
            NrOfCounts[key][0, :]/np.float64(np.sum(NrOfCounts[key][0, :])),
            size=1)

        # Check that not more sample than orignialy existing were present
        ix = new_counts > NrOfCounts[key]
        new_counts[ix] = NrOfCounts[key][ix]

        # If no count is there take the original sample
        if np.sum(new_counts) == 0:
            new_counts = NrOfCounts[key][0, :]

        ix_non_zero = new_counts > 0
        temp_counts = np.zeros((1, np.sum(new_counts > 0)))
        temp_counts[0, :] = new_counts[ix_non_zero]
        NrOfCounts[key] = temp_counts
        Counts[key] = Counts[key][:, ix_non_zero[0, :]]

    return Counts, NrOfCounts




"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    The code in this file has been adapted from the statsmodels package:
    https://github.com/statsmodels/statsmodels/blob/master/statsmodels/genmod/generalized_linear_model.py
    Thus, for this file the licence of the original file applies additionally.
"""

import pickle
import numpy as np
import scipy
from scipy.sparse import csc_matrix, linalg as sla
import statsmodels
from statsmodels.genmod import families
from statsmodels.genmod.generalized_linear_model import GLM
from statsmodels.tools.sm_exceptions import PerfectSeparationError
import time

__all__ = ['GLM']


def _check_convergence(criterion, iteration, tol):
    return not (np.fabs(criterion[iteration] - criterion[iteration-1]) > tol)


class sparse_glm(statsmodels.genmod.generalized_linear_model.GLM):
    def __init__(self, endog, exog, family=None, offset=None, exposure=None,
                 missing='none', data_weights=None, **kwargs):

        if hasattr(family, 'safe_links'):
            if (family is not None) and not isinstance(family.link, tuple(family.safe_links)):
                import warnings
                warnings.warn("The %s link function does not respect the domain of the %s family." %
                              (family.link.__class__.__name__, family.__class__.__name__))
        self.endog = endog
        self.exog = exog
        self.offset = offset
        self.exposure = exposure
        if data_weights is not None:
            self.data_weights = data_weights
        if exposure is not None:
            exposure = np.log(exposure)
        if offset is not None:  # this should probably be done upstream
            offset = np.asarray(offset)

        self._check_inputs(family, self.offset, self.exposure, self.endog)
        if offset is None:
            delattr(self, 'offset')
        if exposure is None:
            delattr(self, 'exposure')

    def fit(self, start_params=None, maxiter=100, method='IRLS', tol=1e-8,
            scale=None, cov_type='nonrobust', cov_kwds=None, use_t=None,
            full_output=True, disp=False, max_start_irls=3, data_weights=None,
            **kwargs):
        """Fits a generalized linear model for a given family.

        parameters
        ----------
        start_params : array-like, optional
            Initial guess of the solution for the loglikelihood maximization.
            The default is family-specific and is given by the
            ``family.starting_mu(endog)``. If start_params is given then the
            initial mean will be calculated as ``np.dot(exog, start_params)``.
        maxiter : int, optional
            Default is 100.
        method : string
            Default is 'IRLS' for iteratively reweighted least squares.
            Otherwise gradient optimization is used.
        tol : float
            Convergence tolerance.  Default is 1e-8.
        scale : string or float, optional
            `scale` can be 'X2', 'dev', or a float
            The default value is None, which uses `X2` for Gamma, Gaussian,
            and Inverse Gaussian.
            `X2` is Pearson's chi-square divided by `df_resid`.
            The default is 1 for the Binomial and Poisson families.
            `dev` is the deviance divided by df_resid
        cov_type : string
            The type of parameter estimate covariance matrix to compute.
        cov_kwds : dict-like
            Extra arguments for calculating the covariance of the parameter
            estimates.
        use_t : bool
            If True, the Student t-distribution is used for inference.
        full_output : bool, optional
            Set to True to have all available output in the Results object's
            mle_retvals attribute. The output is dependent on the solver.
            See LikelihoodModelResults notes section for more information.
            Not used if methhod is IRLS.
        disp : bool, optional
            Set to True to print  convergence messages.  Not used if method is
            IRLS.
        max_start_irls : int
            The number of IRLS iterations used to obtain starting
            values for gradient optimization.  Only relevant if
            `method` is set to something other than 'IRLS'.

        Notes
        -----
        This method does not take any extra undocumented ``kwargs``.
        """
        endog = self.endog
        self.df_resid = self.endog.shape[0] - self.exog.shape[1]

        if not isinstance(data_weights, np.ndarray):
            if endog.ndim > 1 and endog.shape[1] == 2:
                data_weights = endog.sum(1)  # weights are total trials
            else:
                data_weights = np.ones((endog.shape[0], 1))

        self.data_weights = data_weights
        if np.shape(self.data_weights) == () and self.data_weights > 1:
            self.data_weights = self.data_weights * np.ones((endog.shape[0]))
        self.scaletype = scale
        if isinstance(self.family, families.Binomial):
            # This checks what kind of data is given for Binomial.
            # Family will need a reference to endog if this is to be removed
            # from preprocessing
            self.endog = self.family.initialize(self.endog)

        # Construct a combined offset/exposure term.  Note that
        # exposure has already been logged if present.
        offset_exposure = 0.
        if hasattr(self, 'offset'):
            offset_exposure = self.offset
        if hasattr(self, 'exposure'):
            offset_exposure = offset_exposure + self.exposure
        self._offset_exposure = offset_exposure

        if method.lower() == "irls_sparse":
            return self._fit_irls_sparse(
                start_params=start_params, maxiter=maxiter,
                tol=tol, scale=scale, cov_type=cov_type,
                cov_kwds=cov_kwds, use_t=use_t, **kwargs)
        else:
            return self._fit_gradient(
                start_params=start_params,
                method=method,
                maxiter=maxiter,
                tol=tol, scale=scale,
                full_output=full_output,
                disp=disp, cov_type=cov_type,
                cov_kwds=cov_kwds, use_t=use_t,
                max_start_irls=max_start_irls,
                **kwargs)

    def _fit_irls_sparse(
            self, start_params=None, maxiter=50, tol=1e-3, scale=None,
            cov_type='nonrobust', cov_kwds=None, use_t=None, **kwargs):
        """Fit a GLM using IRLS.

        Fit a generalized linear mode (GLM) for a given family using
        iteratively reweighted least squares (IRLS).
        """
        if not scipy.sparse.issparse(self.exog):
            raise ValueError("Matrix not sparse")

        endog = self.endog
        wlsexog = self.exog

        if start_params is None:
            mu = self.family.starting_mu(self.endog)
            lin_pred = self.family.predict(mu)
        else:
            # This is a hack for a faster warm start
            start_params[start_params > 1e2] = 1e2
            start_params[start_params < -1e2] = -1e2
            lin_pred = wlsexog.dot(start_params) + self._offset_exposure
            mu = self.family.fitted(lin_pred)

        dev = self.family.deviance(self.endog[self.endog[:, 0] > 0, :], mu[self.endog[:, 0] > 0, :])
        if np.isnan(dev):
            if not (start_params is None):
                # This is a hack for a faster warm start
                start_params[start_params > 1e1] = 1e1
                start_params[start_params < -1e1] = -1e1
                lin_pred = wlsexog.dot(start_params) + self._offset_exposure
                mu = self.family.fitted(lin_pred)

                dev = self.family.deviance(self.endog, mu)

                if np.isnan(dev):
                    pickle.dump([lin_pred, mu, endog, exog, start_params], open('/home/pdrewe/tmp/tmpdump.' + time.asctime().replace(' ', '_') + '.dat', 'w'))
                    raise ValueError("The first guess on the deviance function "
                                 "returned a nan.  This could be a boundary "
                                 " problem and should be reported.")
            else:
                raise ValueError("The first guess on the deviance function "
                             "returned a nan.  This could be a boundary "
                             " problem and should be reported.")

        # first guess on the deviance is assumed to be scaled by 1.
        # params are none to start, so they line up with the deviance
        history = dict(params=[None, start_params], deviance=[np.inf, dev])
        converged = False
        criterion = history['deviance']
        # This special case is used to get the likelihood for a specific
        # params vector.

        if maxiter == 0:
            pass
        for iteration in range(maxiter):
            self.weights = self.data_weights*self.family.weights(mu)
            wlsendog = (lin_pred + self.family.link.deriv(mu) * (self.endog-mu)
                        - self._offset_exposure)
            W = scipy.sparse.diags(self.weights[:, 0], 0)

            # Compute x for current interation
            temp_mat = wlsexog.transpose().dot(W)
            lu = sla.splu(csc_matrix(temp_mat.dot(wlsexog)))
            wls_results = lu.solve(temp_mat.dot(wlsendog))
            wls_results[wls_results > 1e2] = 1e2
            wls_results[wls_results < -1e2] = -1e2

            lin_pred = self.exog.dot(wls_results) + self._offset_exposure
            mu = self.family.fitted(lin_pred)
            history['mu'] = mu
            history['params'].append(wls_results)
            temp_endog = self.endog[:]
            temp_endog[temp_endog < 0] = 0
            history['deviance'].append(self.family.deviance(self.endog, mu))

            if endog.squeeze().ndim == 1 and np.allclose(mu - endog, 0):
                msg = "Perfect separation detected, results not available"
                raise PerfectSeparationError(msg)
            converged = _check_convergence(criterion, iteration, tol)
            if converged:
                break
        self.mu = mu

        history['iteration'] = iteration + 1

        return [wls_results, history]

    def _check_inputs(self, family, offset, exposure, endog):
        # Default family is Gaussian
        if family is None:
            family = families.Gaussian()
        self.family = family

        if exposure is not None:
            if not isinstance(self.family.link, families.links.Log):
                raise ValueError("exposure can only be used with the log "
                                 "link function")
            elif exposure.shape[0] != endog.shape[0]:
                raise ValueError("exposure is not the same length as endog")

        if offset is not None:
            if offset.shape[0] != endog.shape[0]:
                raise ValueError("offset is not the same length as endog")



"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import resource
import time


def get_mem_usage(verbosity, t=False, msg=''):
    """Return the maximum resident set size with optional message."""
    if verbosity:
        if msg:
            print(msg)
        if t:
            print('Done: Elapsed time: ' + str(time.time() - t))

        mem_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        print('Memory usage: %s (kb)' % mem_usage)



import Cython.Build
from setuptools import setup
from setuptools.extension import Extension

import numpy as np

long_description = "This is a long description"

setup(
    name='omniCLIP_test',
    version="0.1a2",
    description='Test of the omniCLIP project',
    long_description=long_description,
    author='Philipp Boss',
    author_email='joel.simoneau@mdc-berlin.de',
    url='https://github.com/simojoe/omniCLIP',
    cmdclass={'build_ext': Cython.Build.build_ext},
    package_dir={'omniCLIP': 'omniCLIP'},
    packages=['omniCLIP', 'omniCLIP.data_parsing', 'omniCLIP.omni_stat'],
    ext_modules=[Extension(
        'omniCLIP.viterbi',
        sources=['omniCLIP/omni_stat/viterbi.pyx'],
        include_dirs=[np.get_include()],
    )],
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'omniCLIP = omniCLIP.omniCLIP:main'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    setup_requires=['numpy>=1.18', 'cython>=0.24.1'],
    install_requires=[
        'numpy>=1.18', 'nose>=0.11', 'cython>=0.24.1', 'h5py>=2.10.0',
        'statsmodels>=0.11.0', 'scipy>=1.4.1', 'scikit-learn>=0.22.1',
        'pysam>=0.15.3', 'pandas>=1.0.2', 'intervaltree>=3.0.2',
        'gffutils>=0.10.1'],
)



"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from scipy import special
from scipy.sparse import csc_matrix, coo_matrix
from scipy.stats import nbinom
import h5py
import itertools
import multiprocessing
import numpy as np
import scipy as sp
import time

from omniCLIP.omni_stat import sparse_glm
from omniCLIP.omni_stat.utils import get_mem_usage


def NB_parameter_estimation(mean, var):
    """Compute the parameters p and r for the neg-binomial distribution."""
    mean = np.float64(mean)
    var = np.float64(var)

    p = mean / var
    n = (mean ** 2) / (var - mean)

    return p, n


def estimate_expression_param(expr_data, verbosity=1):
    """Estimate the parameters for the expression GLM."""
    (EmissionParameters, Paths) = expr_data

    Sequences = h5py.File(EmissionParameters['dat_file_clip'], 'r')
    Background = h5py.File(EmissionParameters['dat_file_bg'], 'r')

    # 1) Get the library size
    print('Start estimation of expression parameters')
    bg_type = EmissionParameters['bg_type']
    lib_size = EmissionParameters['LibrarySize']
    bck_lib_size = EmissionParameters['BckLibrarySize']
    start_params = EmissionParameters['ExpressionParameters'][0]
    disp = EmissionParameters['ExpressionParameters'][1]

    # 2) Estimate dispersion
    print('Constructing GLM matrix')
    t = time.time()

    # 3) Compute sufficient statistics
    get_mem_usage(
        verbosity,
        msg='Estimating expression parameters: before GLM matrix construction')

    A, w, Y, rep = construct_glm_matrix(
        EmissionParameters, Sequences, Background, Paths)

    print('Estimating expression parameters: GLM matrix constrution')
    get_mem_usage(verbosity, t=t)

    # Make sure that matrix A is in the right format
    if not sp.sparse.isspmatrix_csc(A):
        A = csc_matrix(A)

    get_mem_usage(
        verbosity,
        msg='Estimating expression parameters: before GLM matrix')

    # Create the offset for the library size
    offset = np.zeros_like(rep)
    for i in range(EmissionParameters['NrOfReplicates']):
        offset[rep == (i + 1)] = lib_size[str(i)]
    if bg_type != 'None':
        for i in range(EmissionParameters['NrOfBckReplicates']):
            offset[rep == -(i + 1)] = bck_lib_size[str(i)]

    # 4) Fit GLM
    print('Fitting GLM')
    t = time.time()

    print('Estimating expression parameters: before fitting')
    get_mem_usage(verbosity)

    start_params, disp = fit_glm(
        A, w, Y, offset, disp, start_params,
        norm_class=EmissionParameters['norm_class'])

    get_mem_usage(
        verbosity,
        msg='Estimating expression parameters: after fitting')

    del A, w, Y, offset

    get_mem_usage(
        verbosity, t=t,
        msg='Estimating expression parameters: after cleanup')

    # 5) Process the output
    EmissionParameters['ExpressionParameters'] = [start_params, disp]
    print('Finished expression parameter estimation')

    return EmissionParameters


def construct_glm_matrix(EmissionParameters, Sequences, Background, Paths):
    """Construct the GLM matrix."""
    # Determine shape of the matrix
    bg_type = EmissionParameters['bg_type']
    nr_of_bck_rep = EmissionParameters['NrOfBckReplicates']
    nr_of_rep = EmissionParameters['NrOfReplicates']
    NrOfStates = EmissionParameters['NrOfStates']

    curr_mats = []
    curr_weights = []
    curr_ys = []
    curr_reps = []

    genes = list(Sequences.keys())
    nr_of_genes = len(genes)

    fg_state, bg_state = get_fg_and_bck_state(EmissionParameters)
    # Create a dictionary of unique rows for each gene
    # Check if parallel processing is activated
    replicates_fg = list(Sequences[genes[0]]['Coverage'].keys())
    if EmissionParameters['nb_proc'] == 1:
        for gene_nr, gene in enumerate(Sequences):
            for rep, curr_rep in enumerate(replicates_fg):
                # Create the sparse matrix blocks for each replicate
                curr_data = (
                    Paths[gene], Sequences[gene]['Coverage'][curr_rep][()],
                    gene, gene_nr, rep, NrOfStates, nr_of_genes, bg_type,
                    fg_state, bg_state, EmissionParameters['verbosity'])

                (gene_mat, weights,
                 y, reps, new_pos) = process_gene_for_glm_mat(curr_data)

                del curr_data
                curr_mats.append(gene_mat)
                curr_weights.append(weights)
                curr_ys.append(y)
                curr_reps.append(reps)
    else:
        # Create a function that groups together the values such that an iterator can be defined
        verb = EmissionParameters['verbosity']
        f = lambda gene_nr, gene, rep,  Paths=Paths, Sequences=Sequences, NrOfStates=NrOfStates, nr_of_genes=nr_of_genes, bg_type=bg_type, fg_state=fg_state, bg_state=bg_state, verb=verb: (Paths[gene], Sequences[gene]['Coverage'][str(rep)][()], gene, gene_nr, rep, NrOfStates, nr_of_genes, bg_type, fg_state, bg_state, verb)
        # Create an iterator for the data
        list_gen = [(a, b, c) for (a, b), c in itertools.product(zip(itertools.count(), genes), list(range(nr_of_rep)))]
        data = itertools.starmap(f, list_gen)
        pool = multiprocessing.get_context("spawn").Pool(
            EmissionParameters['nb_proc'], maxtasksperchild=100)
        results = pool.imap(process_gene_for_glm_mat, data, chunksize=1)
        pool.close()
        pool.join()
        results = [res for res in results]
        curr_mats += [res[0] for res in results]
        curr_weights += [res[1] for res in results]
        curr_ys += [res[2] for res in results]
        curr_reps += [res[3] for res in results]
        del results

    # Add row with pseudo counts
    if not bg_type == 'None':
        replicates_bck = list(Background[list(Sequences.keys())[0]]['Coverage'].keys())
        # Process the background
        if EmissionParameters['nb_proc'] == 1:
            for gene_nr, gene in enumerate(Sequences):
                for rep, curr_rep in enumerate(replicates_bck):
                    if bg_type == 'Const':
                        gene_rep_background = Background[gene][curr_rep][()]
                    else:
                        gene_rep_background = Background[gene]['Coverage'][curr_rep][()]

                    curr_data = (Paths[gene], gene_rep_background, gene, gene_nr, rep, NrOfStates, nr_of_genes, bg_type, fg_state, bg_state)
                    gene_mat, weights, y, reps, new_pos = process_bck_gene_for_glm_mat(curr_data)
                    curr_mats.append(gene_mat)
                    curr_weights.append(weights)
                    curr_ys.append(y)
                    curr_reps.append(reps)
        else:
            # Create a function that groups together the values such that an iterator can be defined
            if bg_type == 'Const':
                f = lambda gene_nr, gene, rep,  Paths=Paths, Background=Background, NrOfStates=NrOfStates, nr_of_genes=nr_of_genes, bg_type=bg_type, fg_state=fg_state, bg_state=bg_state: (Paths[gene], Background[gene][str(rep)][()], gene, gene_nr, rep, NrOfStates, nr_of_genes, bg_type, fg_state, bg_state)
            else:
                f = lambda gene_nr, gene, rep,  Paths=Paths, Background=Background, NrOfStates=NrOfStates, nr_of_genes=nr_of_genes, bg_type=bg_type, fg_state=fg_state, bg_state=bg_state: (Paths[gene], Background[gene]['Coverage'][str(rep)][()], gene, gene_nr, rep, NrOfStates, nr_of_genes, bg_type, fg_state, bg_state)

            # Create an iterator for the data
            list_gen = [(a, b, c) for (a, b), c in itertools.product(zip(itertools.count(), list(Sequences.keys())), list(range(nr_of_bck_rep)))]
            data = itertools.starmap(f, list_gen)
            pool = multiprocessing.get_context("spawn").Pool(EmissionParameters['nb_proc'], maxtasksperchild=100)
            results = pool.imap(process_bck_gene_for_glm_mat, data, chunksize=1)
            pool.close()
            pool.join()
            results = [res for res in results]
            curr_mats += [res[0] for res in results]
            curr_weights += [res[1] for res in results]
            curr_ys += [res[2] for res in results]
            curr_reps += [res[3] for res in results]
            del results

    # Remove empty elements
    ix_nz = [curr_mats[i].nnz > 0 for i in range(len(curr_mats))]
    curr_mats = [curr_mats[i] for i in range(len(ix_nz)) if ix_nz[i]]
    curr_weights = [curr_weights[i] for i in range(len(ix_nz)) if ix_nz[i]]
    curr_ys = [curr_ys[i] for i in range(len(ix_nz)) if ix_nz[i]]
    curr_reps = [curr_reps[i] for i in range(len(ix_nz)) if ix_nz[i]]

    # Merge the componetens of all genes
    A = sp.sparse.vstack(curr_mats).tocsc()
    w = np.hstack(curr_weights)
    y = np.hstack(curr_ys)
    rep = np.hstack(curr_reps)

    return A, w, y, rep


def process_bck_gene_for_glm_mat(data):
    """Compute coverage at each position in each state.

    Returns a dictionary with the histogram with of the counts in each state.
    """
    (CurrGenePath, gene_rep_back, gene, gene_nr, rep,
     NrOfStates, nr_of_genes, bg_type, fg_state, bg_state) = data

    # 1) Get the counts
    counts = {}
    if bg_type == 'Const':
        gene_length = len(CurrGenePath)
        counts[gene_rep_back/float(gene_length)] = gene_length
    elif bg_type == 'Coverage' or bg_type == 'Coverage_bck':
        for CurrState in range(NrOfStates):
            curr_counts = gene_rep_back[0, CurrGenePath == CurrState]
            curr_counts[curr_counts < 0] = 0
            counts[CurrState] = np.unique(curr_counts, return_counts=True)
    else:  # 'Count'
        curr_counts = gene_rep_back[0, :]
        curr_counts[curr_counts < 0] = 0
        counts[CurrState] = np.unique(curr_counts, return_counts=True)

    # 2) Generate the matrix for this gene
    curr_y = []
    curr_data = []
    curr_rows = []
    curr_cols = []
    curr_weights = []
    curr_reps = []
    nr_of_rows = 0
    curr_pos = 0

    # Add rows for abundance and weights
    # Estimate the gene expression by a random count
    if bg_type == 'Coverage_bck':
        if np.sum(CurrGenePath == 3) == len(CurrGenePath):
            # Number of tmp samples
            tmp_vals = gene_rep_back[0, :]
            tmp_nb_sample = np.sum(np.sum(tmp_vals > 0))
            if tmp_nb_sample > 0:
                nr_of_sample_sub = min(tmp_nb_sample, 10)
            else:
                nr_of_sample_sub = 10

            curr_y.append(np.random.choice(tmp_vals, nr_of_sample_sub) + 1)

            curr_data.append(np.ones(nr_of_sample_sub))
            curr_rows.append(np.arange(curr_pos, curr_pos + nr_of_sample_sub))
            curr_cols.append(np.ones(nr_of_sample_sub) * gene_nr)
            curr_reps.append(-np.ones(nr_of_sample_sub) * (rep + 1))
            curr_weights.append(np.ones(nr_of_sample_sub))

            curr_pos += nr_of_sample_sub
            nr_of_rows += nr_of_sample_sub

    if bg_type == 'Coverage_bck' or bg_type == 'Coverage':
        for CurrState in range(NrOfStates):
            # Add rows for abundance and weights
            temp_y = counts[CurrState][0]
            temp_weights = counts[CurrState][1]
            curr_weights.append(temp_weights)
            curr_y.append(temp_y)
            nr_of_obs = len(temp_weights)

            curr_data.append(np.ones(nr_of_obs))
            curr_rows.append(np.arange(curr_pos, curr_pos + nr_of_obs))
            # Add information for numper of replicates
            curr_reps.append(-np.ones(nr_of_obs) * (rep + 1))

            if bg_type == 'Coverage':
                curr_cols.append(np.ones(nr_of_obs) * gene_nr)

                if CurrState == fg_state:
                    curr_data.append(-np.ones(nr_of_obs))
                    curr_rows.append(np.arange(curr_pos, curr_pos + nr_of_obs))
                    curr_cols.append(np.ones(nr_of_obs) * (nr_of_genes))
                elif CurrState == bg_state:
                    curr_data.append(np.ones(nr_of_obs))
                    curr_rows.append(np.arange(curr_pos, curr_pos + nr_of_obs))
                    curr_cols.append(np.ones(nr_of_obs) * (nr_of_genes))

            elif bg_type == 'Coverage_bck':
                if CurrState < NrOfStates - 1:
                    curr_cols.append(np.ones(nr_of_obs) * gene_nr)

                    if CurrState == fg_state:
                        curr_data.append(-np.ones(nr_of_obs))
                        curr_rows.append(np.arange(curr_pos, curr_pos + nr_of_obs))
                        curr_cols.append(np.ones(nr_of_obs) * (nr_of_genes))

                    elif CurrState == bg_state:
                        curr_data.append(np.ones(nr_of_obs))
                        curr_rows.append(np.arange(curr_pos, curr_pos + nr_of_obs))
                        curr_cols.append(np.ones(nr_of_obs) * (nr_of_genes))

                else:
                    curr_cols.append(np.ones(nr_of_obs) * (nr_of_genes + 1))

            curr_pos += nr_of_obs
            nr_of_rows += nr_of_obs

    else:
        temp_weights = np.array([counts[c] for c in list(counts.keys())])
        curr_weights.append(temp_weights)
        temp_y = np.array([c for c in list(counts.keys())])
        curr_y.append(temp_y)
        nr_of_obs = len(temp_weights)

        # Add rows for transcript abundance
        curr_data.append(np.ones(nr_of_obs))
        curr_rows.append(np.arange(0, nr_of_obs))
        curr_cols.append(np.ones(nr_of_obs) * gene_nr)

        # Add information for numper of replicates
        curr_reps.append(-np.ones(nr_of_obs) * (rep + 1))
        nr_of_rows += nr_of_obs

    # Create matrix
    if bg_type == 'Coverage_bck':
        shape = (nr_of_rows, nr_of_genes + 2)
    elif bg_type == 'Coverage':
        shape = (nr_of_rows, nr_of_genes + 1)
    else:
        shape = (nr_of_rows, nr_of_genes + NrOfStates)

    gene_mat = coo_matrix(
        (np.hstack(curr_data), (np.hstack(curr_rows), np.hstack(curr_cols))),
        shape=shape)

    weights = np.hstack(curr_weights)
    y = np.hstack(curr_y)
    reps = np.hstack(curr_reps)
    new_pos = curr_pos

    return gene_mat, weights, y, reps, new_pos


def process_gene_for_glm_mat(data):
    """Compute how much coverage is at each position in each state.

    Returns a dictionary with the histogram with of the counts in each state.
    """
    (CurrGenePath, rep_gene_seq, gene, gene_nr, rep, NrOfStates,
     nr_of_genes, bg_type, fg_state, bg_state, verbosity) = data

    # 1) get the counts
    counts = {}
    for CurrState in range(NrOfStates):
        curr_counts = rep_gene_seq[0, CurrGenePath == CurrState]
        curr_counts[curr_counts < 0] = 0
        counts[CurrState] = np.unique(curr_counts, return_counts=True)

    # 2) generate the matrix for this gene
    curr_y = []
    curr_data = []
    curr_rows = []
    curr_cols = []
    curr_weights = []
    curr_reps = []
    nr_of_rows = 0
    curr_pos = 0
    # Check whether all positions are in background state.

    # Estimate the gene expression by a random count
    if bg_type == 'Coverage_bck':
        if np.sum(CurrGenePath == 3) == len(CurrGenePath):
            tmp_vals = rep_gene_seq[0, :]
            tmp_nb_sample = np.sum(np.sum(tmp_vals > 0))
            if tmp_nb_sample > 0:
                tmp_vals = tmp_vals[tmp_vals > 0]
                nr_of_sample_sub = min(tmp_nb_sample, 10)
            else:
                nr_of_sample_sub = 10

            curr_y.append(np.random.choice(tmp_vals, nr_of_sample_sub) + 1)

            curr_data.append(np.ones(nr_of_sample_sub))
            curr_rows.append(np.arange(curr_pos, curr_pos + nr_of_sample_sub))
            curr_cols.append(np.ones(nr_of_sample_sub) * gene_nr)
            curr_reps.append(np.ones(nr_of_sample_sub) * (rep + 1))
            curr_weights.append(np.ones(nr_of_sample_sub))

            curr_pos += nr_of_sample_sub
            nr_of_rows += nr_of_sample_sub
            if verbosity:
                print('estimating gene expression on background for gene: ' + gene)
            # Assume that it is from the state with lower counts

    for CurrState in range(NrOfStates):
        # Add rows for abundance and weights
        temp_y = counts[CurrState][0]
        temp_weights = counts[CurrState][1]
        curr_weights.append(temp_weights)
        curr_y.append(temp_y)

        nr_of_obs = len(temp_weights)
        curr_data.append(np.ones(nr_of_obs))
        curr_rows.append(np.arange(curr_pos, curr_pos + nr_of_obs))
        # Add information for numper of replicates
        curr_reps.append(np.ones(nr_of_obs) * (rep + 1))

        if bg_type == 'None':
            # Add rows for foreground and background
            curr_cols.append(np.ones(nr_of_obs) * CurrState)
        elif bg_type == 'Coverage':
            curr_cols.append(np.ones(nr_of_obs) * gene_nr)

            if CurrState == fg_state:
                curr_data.append(np.ones(nr_of_obs))
                curr_rows.append(np.arange(curr_pos, curr_pos + nr_of_obs))
                curr_cols.append(np.ones(nr_of_obs) * (nr_of_genes))

            elif CurrState == bg_state:
                curr_data.append(-np.ones(nr_of_obs))
                curr_rows.append(np.arange(curr_pos, curr_pos + nr_of_obs))
                curr_cols.append(np.ones(nr_of_obs) * (nr_of_genes))

        elif bg_type == 'Coverage_bck':
            if CurrState < NrOfStates - 1:
                curr_cols.append(np.ones(nr_of_obs) * gene_nr)

                if CurrState == fg_state:
                    curr_data.append(np.ones(nr_of_obs))
                    curr_rows.append(np.arange(curr_pos, curr_pos + nr_of_obs))
                    curr_cols.append(np.ones(nr_of_obs) * (nr_of_genes))

                elif CurrState == bg_state:
                    curr_data.append(-np.ones(nr_of_obs))
                    curr_rows.append(np.arange(curr_pos, curr_pos + nr_of_obs))
                    curr_cols.append(np.ones(nr_of_obs) * (nr_of_genes))

            else:
                curr_cols.append(np.ones(nr_of_obs) * (nr_of_genes + 1))
        else:
            # Add rows for transcript abundance
            curr_cols.append(np.ones(nr_of_obs) * gene_nr)
            # Add rows for foreground and background
            if not((bg_type == 'Coverage_bck') and (CurrState == (NrOfStates - 1))):
                curr_data.append(np.ones(nr_of_obs))
                curr_rows.append(np.arange(curr_pos, curr_pos + nr_of_obs))
                curr_cols.append(np.ones(nr_of_obs) * (nr_of_genes + CurrState))
            else:  # TODO: Code duplication
                curr_data.append(np.ones(nr_of_obs))
                curr_rows.append(np.arange(curr_pos, curr_pos + nr_of_obs))
                curr_cols.append(np.ones(nr_of_obs) * (nr_of_genes + CurrState))

        curr_pos += nr_of_obs
        nr_of_rows += nr_of_obs

    # Create matrix
    if bg_type == 'None':
        shape = (nr_of_rows, NrOfStates)
    elif bg_type == 'Coverage_bck':
        shape = (nr_of_rows, nr_of_genes + 2)
    elif bg_type == 'Coverage':
        shape = (nr_of_rows, nr_of_genes + 1)
    else:  # 'Const' or 'Count'
        shape = (nr_of_rows, nr_of_genes + NrOfStates)

    gene_mat = coo_matrix(
        (np.hstack(curr_data), (np.hstack(curr_rows), np.hstack(curr_cols))),
        shape=shape)

    weights = np.hstack(curr_weights)
    y = np.hstack(curr_y)
    reps = np.hstack(curr_reps)
    new_pos = curr_pos

    return gene_mat, weights, y, reps, new_pos


def fit_glm(A, w, Y, offset, disp=None, start_params=None, norm_class=False,
            verbosity=1):
    """Fit the GLM."""
    if disp is None:
        disp = 1.0
    for i in range(3):
        # Fit GLM
        Y = Y.reshape((len(Y), 1))
        offset = offset.reshape((len(offset), 1))
        w = w.reshape((len(w), 1))
        if disp is not None:
            disp = min(disp, 100)

        tempw = np.float64(w)
        if norm_class:
            Ix_expr = (A[:, -1].toarray() == 1)
            tempw[Ix_expr] = tempw[Ix_expr] / np.sum(tempw[Ix_expr])

            Ix_expr = (A[:, -1].toarray() == 0) * (A[:, -2].toarray() == 0)
            tempw[Ix_expr] = tempw[Ix_expr] / np.sum(tempw[Ix_expr])

            Ix_expr = (A[:, -1].toarray() == 0) * (A[:, -2].toarray() > 0)
            tempw[Ix_expr] = tempw[Ix_expr] / np.sum(tempw[Ix_expr])

            Ix_expr = (A[:, -1].toarray() == 0) * (A[:, -2].toarray() < 0)
            tempw[Ix_expr] = tempw[Ix_expr] / np.sum(tempw[Ix_expr])

        glm_binom = sparse_glm.sparse_glm(
            Y, A, offset=np.log(offset),
            family=sparse_glm.families.NegativeBinomial(alpha=disp))

        res = glm_binom.fit(
            method="irls_sparse", data_weights=tempw,
            start_params=start_params)
        start_params = res[0]

        mu = res[1]['mu']
        del res, glm_binom

        if verbosity > 1:
            print(start_params[-3:])

        disp = sp.optimize.minimize_scalar(
            neg_NB_GLM_loglike,
            bracket=None, bounds=[0.1, 100], args=(Y, mu, tempw,),
            method='bounded', options={'xatol': 0.00001})['x']

        if verbosity > 1:
            print('Dispersion ' + str(disp))

        glm_binom = sparse_glm.sparse_glm(
            Y, A, offset=np.log(offset),
            family=sparse_glm.families.NegativeBinomial(alpha=disp))

        res = glm_binom.fit(
            method="irls_sparse", data_weights=tempw,
            start_params=start_params)

        start_params = res[0]

        if verbosity > 1:
            print((np.sum(np.abs(mu - res[1]['mu']))))
        del glm_binom, res

    # 3) Process the output
    del A, w, Y, offset

    return start_params, disp


def neg_NB_GLM_loglike(alpha, endog, mu, weights):
    """Compute the loglikelihood for a fitted NB GLM."""
    lin_pred = sparse_glm.families.NegativeBinomial(alpha=alpha)._link(mu)
    constant = (special.gammaln(endog + 1 / alpha) -
                special.gammaln(endog+1)-special.gammaln(1/alpha))
    exp_lin_pred = np.exp(lin_pred)

    return -np.sum(
        (endog * np.log(alpha * exp_lin_pred / (1 + alpha * exp_lin_pred))
         - np.log(1 + alpha * exp_lin_pred) / alpha + constant) * weights)


def get_expected_mean_and_var(CurrStackSum, State, nr_of_genes, gene_nr,
                              EmissionParameters, curr_type='fg', verbosity=1):
    """Compute the expected means and variances for each of the states."""
    fg_state, bg_state = get_fg_and_bck_state(EmissionParameters)
    start_params = EmissionParameters['ExpressionParameters'][0]
    disp = EmissionParameters['ExpressionParameters'][1]
    nr_of_states = EmissionParameters['NrOfStates']
    # Create the log mean matrix for each rep

    bg_type = EmissionParameters['bg_type']
    if curr_type == 'fg':
        lib_size = EmissionParameters['LibrarySize']
        nr_of_rep = EmissionParameters['NrOfReplicates']
    else:
        lib_size = EmissionParameters['BckLibrarySize']
        nr_of_rep = EmissionParameters['NrOfBckReplicates']

    gene_len = CurrStackSum.shape[1]

    GLM_mat = np.zeros((nr_of_rep, 1))

    if bg_type == 'Coverage_bck':
        start_params = (start_params)
        if start_params[gene_nr, 0] - abs(start_params[nr_of_genes, 0]) < start_params[nr_of_genes + 1, 0]:
            start_params[gene_nr, 0] = abs(start_params[nr_of_genes, 0]) + start_params[nr_of_genes + 1, 0] + 0.01
            if EmissionParameters['verbosity']:
                print('Adjusted start params')

    # Check whether the current data is foreground data
    for rep in range(nr_of_rep):
        GLM_mat[rep, :] += np.ones((1)) * np.log(lib_size[str(rep)])
        if curr_type == 'fg':
            if bg_type == 'None':
                GLM_mat[rep, :] += np.ones((1)) * start_params[State, 0]
            elif bg_type == 'Coverage_bck':
                if State < nr_of_states - 1:
                    GLM_mat[rep, :] += np.ones((1)) * start_params[gene_nr, 0]
                    if State == fg_state:  # Skip the last state
                        GLM_mat[rep, :] += np.ones((1)) * start_params[nr_of_genes, 0]
                    elif State == bg_state:  # TODO: Code duplication
                        GLM_mat[rep, :] -= np.ones((1)) * start_params[nr_of_genes, 0]
                    else:
                        continue
                else:
                    GLM_mat[rep, :] += np.ones((1)) * start_params[nr_of_genes + 1, 0]

            elif bg_type == 'Coverage':
                GLM_mat[rep, :] += np.ones((1)) * start_params[gene_nr, 0]
                if State == fg_state:  # Skip the last state
                    GLM_mat[rep, :] += np.ones((1)) * start_params[nr_of_genes, 0]
                elif State == bg_state:  # TODO: Code duplication
                    GLM_mat[rep, :] -= np.ones((1)) * start_params[nr_of_genes, 0]
                else:
                    continue
            else:
                GLM_mat[rep, :] += np.ones((1)) * start_params[gene_nr, 0]
                GLM_mat[rep, :] += np.ones((1)) * start_params[nr_of_genes + State, 0]
        else:
            if bg_type == 'None':
                GLM_mat[rep, :] += np.ones((1)) * start_params[State, 0]
            elif bg_type == 'Coverage_bck':
                if State < nr_of_states - 1:
                    GLM_mat[rep, :] += np.ones((1)) * start_params[gene_nr, 0]
                    if State == fg_state:  # Skip the last state
                        GLM_mat[rep, :] -= np.ones((1)) * start_params[nr_of_genes, 0]
                    elif State == bg_state:  # TODO: Code duplication
                        GLM_mat[rep, :] += np.ones((1)) * start_params[nr_of_genes, 0]
                    else:
                        continue
                else:
                    GLM_mat[rep, :] += np.ones((1)) * start_params[nr_of_genes + 1, 0]
            elif bg_type == 'Coverage':
                GLM_mat[rep, :] += np.ones((1)) * start_params[gene_nr, 0]
                if State == fg_state:  # Skip the last state
                    GLM_mat[rep, :] -= np.ones((1)) * start_params[nr_of_genes, 0]
                elif State == bg_state:  # TODO: Code duplication
                    GLM_mat[rep, :] += np.ones((1)) * start_params[nr_of_genes, 0]
                else:
                    continue
            else:
                GLM_mat[rep, :] += np.ones((1)) * start_params[gene_nr, 0]

    # Compute the mean of the positions in each state
    mean_mat = np.tile(np.exp(GLM_mat), (1, gene_len))

    # Compute the variance of the position
    var_mat = mean_mat + disp * (mean_mat ** 2)

    # Adjust variance for the background state
    if State == 3:
        if bg_type == 'Coverage_bck':
            GLM_mat_alt = np.zeros((nr_of_rep, 1))
            for rep in range(nr_of_rep):
                GLM_mat_alt[rep, :] += np.ones((1)) * np.log(lib_size[str(rep)])
            GLM_mat_alt += start_params[gene_nr, 0] - abs(start_params[nr_of_genes, 0])
            mean_mat_alt = np.tile(np.exp(GLM_mat_alt), (1, gene_len))
            disp_alt = mean_mat_alt * disp / mean_mat
            var_mat_alt = mean_mat * disp_alt * (mean_mat ** 2)
            ix = var_mat_alt > var_mat
            var_mat[ix] = var_mat_alt[ix]

    # make sure that var >mean
    ix = (mean_mat * 1.000001) > var_mat
    var_mat[ix] = mean_mat[ix] * 1.000001

    return mean_mat, var_mat


def predict_expression_log_likelihood_for_gene(
        CurrStackSum, State, nr_of_genes, gene_nr, EmissionParameters,
        curr_type='fg'):
    """Predict the likelihood of expression for a gene."""
    nr_of_rep = EmissionParameters['NrOfReplicates']
    mean_mat, var_mat = get_expected_mean_and_var(
        CurrStackSum, State, nr_of_genes, gene_nr,
        EmissionParameters, curr_type=curr_type)

    # Reset the  variance if the empirical variance is larger than the
    # estimated variance
    if EmissionParameters['emp_var'] and (mean_mat.shape[0] > 1):
        mu = np.zeros_like(mean_mat)
        emp_var = np.zeros_like(mean_mat)
        for rep in range(nr_of_rep):
            for i in range(nr_of_rep):
                mu[i, :] = CurrStackSum[i, :] / mean_mat[i, :] * mean_mat[rep, :]
            emp_var[rep, :] = np.var(mu, axis=0)
            var_mat[rep, :] = np.max(np.vstack([var_mat[rep, :], emp_var[rep, :]]), axis=0)

    # Get the parameters for the NB distributions
    p, n = NB_parameter_estimation(mean_mat, var_mat)

    # Compute the likelihood
    ix_nonzero = np.sum(CurrStackSum, axis=0) > 0
    zero_array = nbinom._logpmf(
        np.zeros((CurrStackSum.shape[0], 1)),
        np.expand_dims(n[:, 0], axis=1),
        np.expand_dims(p[:, 0], axis=1))

    loglike = np.tile(zero_array, (1, CurrStackSum.shape[1]))
    loglike[:, ix_nonzero] = nbinom._logpmf(
        CurrStackSum[:, ix_nonzero],
        n[:, ix_nonzero],
        p[:, ix_nonzero])

    # Combine the replicates
    loglike = loglike.sum(axis=0)

    return loglike


def get_fg_and_bck_state(EmissionParameters, final_pred=False):
    """Determine the foreground and background state."""
    # Check whether the parameters have been already defined
    if final_pred:
        if EmissionParameters['ExpressionParameters'][0] is None:
            fg_state = 0
            bg_state = 1
        else:
            bg_type = EmissionParameters['bg_type']
            if bg_type == 'Coverage_bck':
                if EmissionParameters['ExpressionParameters'][0][-2] > 0:
                    fg_state = 1
                    bg_state = 0
                else:
                    fg_state = 0
                    bg_state = 1
            else:
                if EmissionParameters['ExpressionParameters'][0][-1] > 0:
                    fg_state = 1
                    bg_state = 0
                else:
                    fg_state = 0
                    bg_state = 1
    else:
        fg_state = 1
        bg_state = 0

    return fg_state, bg_state
